#
# Copyright (c) 2021, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# Adapted from source code: https://github.com/karlhigley/ranking-metrics-torch
from typing import List, Optional, Sequence

import tensorflow as tf
from keras.utils import losses_utils, metrics_utils
from keras.utils.tf_utils import is_tensor_or_variable
from tensorflow.keras import backend
from tensorflow.keras.metrics import Mean
from tensorflow.keras.metrics import get as get_metric

from merlin.models.tf.metrics import metrics_registry
from merlin.models.tf.utils.tf_utils import extract_topk

METRIC_PARAMETERS_DOCSTRING = """
    y_true : tf.Tensor
        A tensor with shape (batch_size, n_items) corresponding to
        the multi-hot representation of true labels.
    y_pred : tf.Tensor
        A tensor with shape (batch_size, n_items) corresponding to
        the prediction scores.
    label_relevant_counts: tf.Tensor
        A 1D tensor (batch size) which contains the total number of relevant items
        for the example. This is necessary when setting `pre_sorting=True` on the
        ranking metrics classes (e.g. RecallAt(5, pre_sorted=True)), as extract_topk
        is used to extract only the top-k predictions and corresponding labels,
        potentially losing other relevant items not among top-k. In such cases,
        the label_relevant_counts will contain the total relevant counts per example.
    k : int
        The cut-off for ranking metrics
"""


def recall_at(
    y_true: tf.Tensor,
    y_pred: tf.Tensor,
    label_relevant_counts: Optional[tf.Tensor] = None,
    k: int = 5,
) -> tf.Tensor:
    """
    Computes Recall@K metric
    ----------
    {METRIC_PARAMETERS_DOCSTRING}
    """
    rel_count = tf.clip_by_value(label_relevant_counts, clip_value_min=1, clip_value_max=float(k))

    rel_labels = tf.reduce_sum(y_true[:, :k], axis=-1)
    results = tf.cast(
        tf.math.divide_no_nan(rel_labels, rel_count),
        backend.floatx(),
    )
    return results


def precision_at(
    y_true: tf.Tensor,
    y_pred: tf.Tensor,
    label_relevant_counts: Optional[tf.Tensor] = None,
    k: int = 5,
) -> tf.Tensor:
    """
    Computes Precision@K metric
    Parameters
    ----------
    {METRIC_PARAMETERS_DOCSTRING}
    """
    results = tf.cast(tf.reduce_mean(y_true[:, : int(k)], axis=-1), backend.floatx())
    return results


def average_precision_at(
    y_true: tf.Tensor,
    y_pred: tf.Tensor,
    label_relevant_counts: Optional[tf.Tensor] = None,
    k: int = 5,
) -> tf.Tensor:
    """
    Computes Mean Average Precision (MAP) @K
    Parameters
    ----------
    {METRIC_PARAMETERS_DOCSTRING}
    """
    # Computing the precision from 1 to k range
    precisions = tf.stack([precision_at(y_true, y_pred, k=_k) for _k in range(1, k + 1)], axis=-1)
    # Keeping only the precision at the position of relevant items
    rel_precisions = precisions * y_true[:, :k]

    total_prec = tf.reduce_sum(rel_precisions, axis=-1)
    total_relevant_topk = tf.clip_by_value(
        label_relevant_counts, clip_value_min=1, clip_value_max=float(k)
    )

    results = tf.cast(tf.math.divide_no_nan(total_prec, total_relevant_topk), backend.floatx())

    return results


def dcg_at(
    y_true: tf.Tensor,
    y_pred: tf.Tensor,
    label_relevant_counts: Optional[tf.Tensor] = None,
    k: int = 5,
    log_base: int = 2,
) -> tf.Tensor:

    """
    Compute discounted cumulative gain @K (ignoring ties)
    Parameters
    ----------
    {METRIC_PARAMETERS_DOCSTRING}
    """

    # Compute discounts
    discount_positions = tf.cast(tf.range(k), backend.floatx())
    discount_log_base = tf.math.log(tf.convert_to_tensor([log_base], dtype=backend.floatx()))

    discounts = 1 / (tf.math.log(discount_positions + 2) / discount_log_base)
    m = y_true[:, :k] * tf.repeat(tf.expand_dims(discounts[:k], 0), tf.shape(y_true)[0], axis=0)

    results = tf.cast(tf.reduce_sum(m, axis=-1), backend.floatx())
    return results


def ndcg_at(
    y_true: tf.Tensor,
    y_pred: tf.Tensor,
    label_relevant_counts: Optional[tf.Tensor] = None,
    k: int = 5,
    log_base: int = 2,
) -> tf.Tensor:
    """
    Compute normalized discounted cumulative gain @K (ignoring ties)
    Parameters
    ----------
    {METRIC_PARAMETERS_DOCSTRING}
    log_base : int
        Base of the log discount where relevant items are rankied. Defaults to 2
    """
    gains = dcg_at(y_true, y_pred, k=k, log_base=log_base)
    perfect_labels_sorting = tf.cast(
        tf.cast(tf.expand_dims(tf.range(k), 0), label_relevant_counts.dtype)  # type: ignore
        < tf.expand_dims(label_relevant_counts, -1),
        backend.floatx(),
    )
    ideal_gains = dcg_at(perfect_labels_sorting, perfect_labels_sorting, k=k, log_base=log_base)

    results = tf.cast(tf.math.divide_no_nan(gains, ideal_gains), backend.floatx())
    return results


def mrr_at(
    y_true: tf.Tensor,
    y_pred: tf.Tensor,
    label_relevant_counts: Optional[tf.Tensor] = None,
    k: int = 5,
) -> tf.Tensor:
    """
    Compute MRR
    ----------
    {METRIC_PARAMETERS_DOCSTRING}
    """

    first_rel_position = tf.cast(tf.argmax(y_true, axis=-1) + 1, backend.floatx())
    relevant_mask = tf.reduce_max(y_true[:, : int(k)], axis=-1)

    rel_position = first_rel_position * relevant_mask
    results = tf.cast(tf.math.divide_no_nan(1.0, rel_position), backend.floatx())

    return results


@tf.keras.utils.register_keras_serializable(package="merlin.models")
class RankingMetric(Mean):
    def __init__(self, fn, k=5, pre_sorted=True, name=None, dtype=None, **kwargs):
        if name is not None:
            name = f"{name}_{k}"
        super().__init__(name=name, dtype=dtype)
        self._fn = fn
        self.k = k
        self.pre_sorted = pre_sorted
        self._fn_kwargs = kwargs

    def update_state(
        self,
        y_true: tf.Tensor,
        y_pred: tf.Tensor,
        label_relevant_counts: Optional[tf.Tensor] = None,
        sample_weight: Optional[tf.Tensor] = None,
    ):
        y_true, y_pred = self.check_cast_inputs(y_true, y_pred)
        (
            [
                y_true,
                y_pred,
            ],
            sample_weight,
        ) = metrics_utils.ragged_assert_compatible_and_get_flat_values(
            [y_true, y_pred], sample_weight
        )
        y_pred, y_true = losses_utils.squeeze_or_expand_dimensions(y_pred, y_true)

        tf.debugging.assert_greater_equal(
            tf.shape(y_true)[1],
            self.k,
            f"The ranking metric ({self.name}) cutoff ({self.k}) cannot be smaller than "
            f"the number of predictions per example",
        )

        y_pred, y_true, label_relevant_counts = self._maybe_sort_top_k(
            y_pred, y_true, label_relevant_counts
        )

        ag_fn = tf.__internal__.autograph.tf_convert(
            self._fn, tf.__internal__.autograph.control_status_ctx()
        )

        matches = ag_fn(
            y_true,
            y_pred,
            label_relevant_counts=label_relevant_counts,
            k=self.k,
            **self._fn_kwargs,
        )
        return super().update_state(matches, sample_weight=sample_weight)

    def _maybe_sort_top_k(self, y_pred, y_true, label_relevant_counts: tf.Tensor = None):
        if not self.pre_sorted:
            y_pred, y_true, label_relevant_counts = extract_topk(self.k, y_pred, y_true)
        else:
            if label_relevant_counts is None:
                raise Exception(
                    "If y_true was pre-sorted (and truncated to top-k) you must "
                    "provide label_relevant_counts argument."
                )
            label_relevant_counts = tf.cast(label_relevant_counts, self._dtype)

        return y_pred, y_true, label_relevant_counts

    def check_cast_inputs(self, labels, predictions):
        tf.assert_equal(
            tf.rank(predictions), 2, f"predictions must be 2-D tensor (got {predictions.shape})"
        )
        tf.assert_equal(tf.rank(labels), 2, f"labels must be 2-D tensor (got {labels.shape})")

        predictions.get_shape().assert_is_compatible_with(labels.get_shape())

        return tf.cast(labels, self._dtype), tf.cast(predictions, self._dtype)

    def get_config(self):
        config = {}

        if type(self) is RankingMetric:
            # Only include function argument when the object is of a subclass.
            config["fn"] = self._fn
            config["k"] = self.k
            config["pre_sorted"] = self.pre_sorted

            for k, v in self._fn_kwargs.items():
                config[k] = backend.eval(v) if is_tensor_or_variable(v) else v
            base_config = super(RankingMetric, self).get_config()
            return dict(list(base_config.items()) + list(config.items()))

        return {}

    @classmethod
    def from_config(cls, config):
        fn = config.pop("fn", None)
        k = config.pop("k", None)
        pre_sorted = config.pop("pre_sorted", None)
        if cls is RankingMetric:
            return cls(get_metric(fn), k=k, pre_sorted=pre_sorted, **config)
        return super(RankingMetric, cls).from_config(config)


@metrics_registry.register_with_multiple_names("recall_at", "recall")
class RecallAt(RankingMetric):
    def __init__(self, k=10, pre_sorted=False, name="recall_at"):
        super().__init__(recall_at, k=k, pre_sorted=pre_sorted, name=name)


@metrics_registry.register_with_multiple_names("precision_at", "precision")
class PrecisionAt(RankingMetric):
    def __init__(self, k=10, pre_sorted=False, name="precision_at"):
        super().__init__(precision_at, k=k, pre_sorted=pre_sorted, name=name)


@metrics_registry.register_with_multiple_names("map_at", "map")
class AvgPrecisionAt(RankingMetric):
    def __init__(self, k=10, pre_sorted=False, name="map_at"):
        super().__init__(average_precision_at, k=k, pre_sorted=pre_sorted, name=name)


@metrics_registry.register_with_multiple_names("mrr_at", "mrr")
class MRRAt(RankingMetric):
    def __init__(self, k=10, pre_sorted=False, name="mrr_at"):
        super().__init__(mrr_at, k=k, pre_sorted=pre_sorted, name=name)


@metrics_registry.register_with_multiple_names("ndcg_at", "ndcg")
class NDCGAt(RankingMetric):
    def __init__(self, k=10, pre_sorted=False, name="ndcg_at"):
        super().__init__(ndcg_at, k=k, pre_sorted=pre_sorted, name=name)


def ranking_metrics(top_ks: Sequence[int], **kwargs) -> Sequence[RankingMetric]:
    metrics: List[RankingMetric] = []
    for k in top_ks:
        metrics.extend([RecallAt(k), MRRAt(k), NDCGAt(k), AvgPrecisionAt(k), PrecisionAt(k)])
    return metrics



#
# Copyright (c) 2021, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from typing import Dict, Optional

import torch

from merlin.models.torch.features.embedding import FeatureConfig, TableConfig
from merlin.models.torch.tabular.base import TabularTransformation, tabular_transformation_registry
from merlin.models.torch.typing import TabularData, TensorOrTabularData


@tabular_transformation_registry.register_with_multiple_names("stochastic-swap-noise", "ssn")
class StochasticSwapNoise(TabularTransformation):
    """
    Applies Stochastic replacement of sequence features.
    It can be applied as a `pre` transform like `TransformerBlock(pre="stochastic-swap-noise")`
    """

    def __init__(self, schema=None, pad_token=0, replacement_prob=0.1):
        super().__init__()
        self.schema = schema
        self.pad_token = pad_token
        self.replacement_prob = replacement_prob

    def forward(  # type: ignore
        self, inputs: TensorOrTabularData, input_mask: Optional[torch.Tensor] = None, **kwargs
    ) -> TensorOrTabularData:
        if self._schema:
            input_mask = input_mask or self.get_padding_mask_from_item_id(inputs, self.pad_token)
        if isinstance(inputs, dict):
            return {key: self.augment(val, input_mask) for key, val in inputs.items()}

        return self.augment(inputs, input_mask)

    def forward_output_size(self, input_size):
        return input_size

    def augment(
        self, input_tensor: torch.Tensor, mask: Optional[torch.Tensor] = None
    ) -> torch.Tensor:
        with torch.no_grad():
            if mask is not None:
                if input_tensor.ndim == mask.ndim - 1:
                    mask = mask[:, 0]

            sse_prob_replacement_matrix = torch.full(
                input_tensor.shape,
                self.replacement_prob,
                device=input_tensor.device,
            )
            sse_replacement_mask = torch.bernoulli(sse_prob_replacement_matrix).bool()
            if mask is not None:
                sse_replacement_mask = sse_replacement_mask & mask
            n_values_to_replace = sse_replacement_mask.sum()

            if mask is not None:
                masked = torch.masked_select(input_tensor, mask)
            else:
                masked = torch.clone(input_tensor)

            input_permutation = torch.randperm(masked.shape[0])
            sampled_values_to_replace = masked[input_permutation][
                :n_values_to_replace  # type: ignore
            ]
            output_tensor = input_tensor.clone()

            if input_tensor[sse_replacement_mask].size() != sampled_values_to_replace:
                sampled_values_to_replace = torch.squeeze(sampled_values_to_replace)

            output_tensor[sse_replacement_mask] = sampled_values_to_replace

        return output_tensor


@tabular_transformation_registry.register_with_multiple_names("layer-norm")
class TabularLayerNorm(TabularTransformation):
    """
    Applies Layer norm to each input feature individually, before the aggregation
    """

    def __init__(self, features_dim: Optional[Dict[str, int]] = None):
        super().__init__()
        feature_layer_norm = {}
        if features_dim:
            for fname, dim in features_dim.items():
                feature_layer_norm[fname] = torch.nn.LayerNorm(normalized_shape=dim)
        self.feature_layer_norm = torch.nn.ModuleDict(feature_layer_norm)

    @classmethod
    def from_feature_config(cls, feature_config: Dict[str, FeatureConfig]):
        features_dim = {}
        for name, feature in feature_config.items():
            table: TableConfig = feature.table
            features_dim[name] = table.dim
        return cls(features_dim)

    def forward(self, inputs: TabularData, **kwargs) -> TabularData:
        if len(self.feature_layer_norm) == 0:
            raise ValueError("`features_dim is empty.`")
        return {key: self.feature_layer_norm[key](val) for key, val in inputs.items()}

    def forward_output_size(self, input_size):
        return input_size

    def build(self, input_size, **kwargs):
        if len(self.feature_layer_norm) == 0:
            for key, size in input_size.items():
                self.feature_layer_norm[key] = torch.nn.LayerNorm(normalized_shape=size[-1])

        return super().build(input_size, **kwargs)



#
# Copyright (c) 2021, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import merlin.models.torch as ml
from merlin.schema import Tags


def test_continuous_features(torch_con_features):
    features = ["con_a", "con_b"]
    con = ml.ContinuousFeatures(features)(torch_con_features)

    assert list(con.keys()) == features


def test_continuous_features_yoochoose(tabular_schema, torch_tabular_data):
    cont_cols = tabular_schema.select_by_tag(Tags.CONTINUOUS)

    con = ml.ContinuousFeatures.from_schema(cont_cols)
    outputs = con(torch_tabular_data)

    assert set(outputs.keys()) == set(cont_cols.column_names)



#
# Copyright (c) 2021, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import pytest
import torch

import merlin.models.torch as ml


@pytest.mark.parametrize("task", [ml.BinaryClassificationTask, ml.RegressionTask])
def test_simple_heads(torch_tabular_features, torch_tabular_data, task):
    targets = {"target": torch.randint(2, (100,)).float()}

    body = ml.SequentialBlock(torch_tabular_features, ml.MLPBlock([64]))
    head = task("target").to_head(body, torch_tabular_features)

    body_out = body(torch_tabular_data)
    loss = head.compute_loss(body_out, targets)

    assert loss.min() >= 0 and loss.max() <= 1


@pytest.mark.parametrize("task", [ml.BinaryClassificationTask, ml.RegressionTask])
@pytest.mark.parametrize("task_block", [None, ml.MLPBlock([32]), ml.MLPBlock([32]).build([-1, 64])])
# @pytest.mark.parametrize("summary", ["last", "first", "mean", "cls_index"])
def test_simple_heads_on_sequence(torch_tabular_features, torch_tabular_data, task, task_block):
    inputs = torch_tabular_features
    targets = {"target": torch.randint(2, (100,)).float()}

    body = ml.SequentialBlock(inputs, ml.MLPBlock([64]))
    head = task("target", task_block=task_block).to_head(body, inputs)

    body_out = body(torch_tabular_data)
    loss = head.compute_loss(body_out, targets)

    assert loss.min() >= 0 and loss.max() <= 1


@pytest.mark.parametrize(
    "task_blocks",
    [
        None,
        ml.MLPBlock([32]),
        ml.MLPBlock([32]).build([-1, 64]),
        dict(classification=ml.MLPBlock([16]), regression=ml.MLPBlock([20])),
    ],
)
def test_head_with_multiple_tasks(torch_tabular_features, torch_tabular_data, task_blocks):
    targets = {
        "classification": torch.randint(2, (100,)).float(),
        "regression": torch.randint(2, (100,)).float(),
    }

    body = ml.SequentialBlock(torch_tabular_features, ml.MLPBlock([64]))
    tasks = [
        ml.BinaryClassificationTask("classification", task_name="classification"),
        ml.RegressionTask("regression", task_name="regression"),
    ]
    head = ml.Head(body, tasks, task_blocks=task_blocks)
    optimizer = torch.optim.Adam(head.parameters())

    with torch.set_grad_enabled(mode=True):
        body_out = body(torch_tabular_data)
        loss = head.compute_loss(body_out, targets)
        metrics = head.calculate_metrics(body_out, targets, call_body=False)

        optimizer.zero_grad()
        loss.backward()
        optimizer.step()

    assert loss.min() >= 0 and loss.max() <= 1
    assert len(metrics.keys()) == 4
    if task_blocks:
        assert head.task_blocks["classification"][0] != head.task_blocks["regression"][0]

        assert not torch.equal(
            head.task_blocks["classification"][0][0].weight,
            head.task_blocks["regression"][0][0].weight,
        )



import copy
import sys
from functools import reduce
from typing import Dict, List, Optional, Union

import six
import tensorflow as tf

from merlin.models.tf.blocks.core.base import (
    Block,
    BlockType,
    NoOp,
    PredictionOutput,
    is_input_block,
    right_shift_layer,
)
from merlin.models.tf.blocks.core.tabular import Filter, TabularAggregationType, TabularBlock
from merlin.models.tf.utils import tf_utils
from merlin.models.tf.utils.tf_utils import call_layer
from merlin.models.utils import schema_utils
from merlin.schema import Schema, Tags


@tf.keras.utils.register_keras_serializable(package="merlin.models")
class SequentialBlock(Block):
    """The SequentialLayer represents a sequence of Keras layers.
    It is a Keras Layer that can be used instead of tf.keras.layers.Sequential,
    which is actually a Keras Model.  In contrast to keras Sequential, this
    layer can be used as a pure Layer in tf.functions and when exporting
    SavedModels, without having to pre-declare input and output shapes.  In turn,
    this layer is usable as a preprocessing layer for TF Agents Networks, and
    can be exported via PolicySaver.
    Usage::

        c = SequentialLayer([layer1, layer2, layer3])
        output = c(inputs)    # Equivalent to: output = layer3(layer2(layer1(inputs)))
    """

    def __init__(
        self,
        *layers,
        filter: Optional[Union[Schema, Tags, List[str], "Filter"]] = None,
        pre_aggregation: Optional["TabularAggregationType"] = None,
        block_name: Optional[str] = None,
        copy_layers: bool = False,
        **kwargs,
    ):
        """Create a sequential composition.

        Parameters
        ----------
        layers:
            A list or tuple of layers to compose.
        **kwargs:
            Arguments to pass to `Keras` layer initializer, including `name`.

        Raises
        ------
        TypeError:
            If any of the layers are not instances of keras `Layer`.
        """
        if len(layers) == 1 and isinstance(layers[0], (list, tuple)):
            layers = layers[0]  # type: ignore

        self.block_name = block_name

        if pre_aggregation:
            layers = [TabularBlock(aggregation=pre_aggregation), *layers]  # type: ignore

        for layer in layers:
            if not isinstance(layer, tf.keras.layers.Layer):
                raise TypeError(
                    "Expected all layers to be instances of keras Layer, but saw: '{}'".format(
                        layer
                    )
                )

        super(SequentialBlock, self).__init__(**kwargs)

        if getattr(layers[0], "has_schema", None):
            super().set_schema(layers[0].schema)

            for layer in layers[1:]:
                if hasattr(layer, "set_schema"):
                    layer.set_schema(layers[0].schema)

        layers = copy.copy(layers) if copy_layers else layers
        if filter:
            if not isinstance(filter, Filter):
                filter = Filter(filter)
            self.layers = [filter, *layers]
        else:
            self.layers = list(layers)

    def compute_output_shape(self, input_shape):
        """Computes the output shape based on the input shape

        Parameters
        ----------
        input_shape : tf.TensorShape
            The input shape

        Returns
        -------
        tf.TensorShape
            The output shape
        """
        output_shape = input_shape
        for layer in self.layers:
            output_shape = layer.compute_output_shape(output_shape)
        return output_shape

    def compute_output_signature(self, input_signature):
        output_signature = input_signature
        for layer in self.layers:
            output_signature = layer.compute_output_signature(output_signature)
        return output_signature

    def build(self, input_shape=None):
        """Builds the sequential block

        Parameters
        ----------
        input_shape : tf.TensorShape, optional
            The input shape, by default None
        """
        self._maybe_propagate_context(input_shape)
        last_layer = None
        for layer in self.layers:
            try:
                layer.build(input_shape)
            except TypeError:
                t, v, tb = sys.exc_info()
                if isinstance(input_shape, dict) and isinstance(last_layer, TabularBlock):
                    v = TypeError(
                        f"Couldn't build {layer}, "
                        f"did you forget to add aggregation to {last_layer}?"
                    )
                six.reraise(t, v, tb)
            input_shape = layer.compute_output_shape(input_shape)
            last_layer = layer
        self.built = True

    def set_schema(self, schema=None):
        for layer in self.layers:
            self._maybe_set_schema(layer, schema)

        return super().set_schema(schema)

    def _get_name(self):
        return self.block_name if self.block_name else f"{self.__class__.__name__}"

    @property
    def inputs(self):
        """Returns the InputBlock, if it is the first
        block within SequenceBlock

        Returns
        -------
        InputBlock
            The input block
        """
        first = list(self)[0]
        if isinstance(first, SequentialBlock):
            return first.inputs
        if is_input_block(first):
            return first

    @property
    def first(self):
        """Returns the first block in the SequenceBlock

        Returns
        -------
        Block
            The first block of SequenceBlock
        """
        return self.layers[0]

    @property
    def last(self):
        """Returns the last block in the SequenceBlock

        Returns
        -------
        Block
            The last block of SequenceBlock
        """
        return self.layers[-1]

    @property
    def filter_features(self) -> List[str]:
        if isinstance(self.layers[0], Filter):
            return self.layers[0].feature_names
        elif isinstance(self.layers[0], SequentialBlock):
            return self.layers[0].filter_features

        return []

    @property
    def trainable_weights(self):
        """Returns trainable weights of all layers
        of this block

        Returns
        -------
        List
            List with trainable weights
        """
        if not self.trainable:
            return []
        weights = {}
        for layer in self.layers:
            for v in layer.trainable_weights:
                weights[id(v)] = v
        return list(weights.values())

    @property
    def non_trainable_weights(self):
        """Returns non-trainable weights of all layers
        of this block

        Returns
        -------
        List
            List with non-trainable weights
        """
        weights = {}
        for layer in self.layers:
            for v in layer.non_trainable_weights:
                weights[id(v)] = v
        return list(weights.values())

    @property
    def trainable(self):
        """Returns whether all layer within SequentialBlock are trainable

        Returns
        -------
        bool
            True if all layer within SequentialBlock are trainable, otherwise False
        """
        return all(layer.trainable for layer in self.layers)

    @trainable.setter
    def trainable(self, value):
        """Makes all block layers trainable or not

        Parameters
        ----------
        value : bool
            Sets all layers trainable flag
        """
        for layer in self.layers:
            layer.trainable = value

    @property
    def losses(self):
        values, _val_names = [], set()
        for layer in self.layers:
            losses = layer.losses
            for loss in losses:
                if isinstance(loss, tf.Tensor):
                    if loss.ref() not in _val_names:
                        _val_names.add(loss.ref())
                        values.append(loss)
                    else:
                        raise ValueError(f"Loss should be a Tensor, found: {loss}")

        return values

    @property
    def regularizers(self):
        values = set()
        for layer in self.layers:
            values.update(layer.regularizers)
        return list(values)

    def call(self, inputs, training=False, **kwargs):
        outputs = inputs
        for i, layer in enumerate(self.layers):
            outputs = call_layer(layer, outputs, training=training, **kwargs)

        return outputs

    def compute_loss(self, inputs, targets, **kwargs):
        outputs, targets = inputs, targets
        for layer in self.layers:
            outputs, targets = layer.compute_loss(outputs, targets=targets, **kwargs)

        return outputs, targets

    def call_outputs(
        self, outputs: PredictionOutput, training=False, **kwargs
    ) -> "PredictionOutput":
        for layer in self.layers:
            outputs = layer.call_outputs(outputs, training=training, **kwargs)
        return outputs

    def get_config(self):
        config = {}
        for i, layer in enumerate(self.layers):
            config[i] = tf.keras.utils.serialize_keras_object(layer)

        return config

    def __getitem__(self, key):
        return self.layers[key]

    @property
    def is_tabular(self):
        return getattr(self.layers[-1], "is_tabular", False)

    @classmethod
    def from_config(cls, config, custom_objects=None):
        layers = [
            tf.keras.layers.deserialize(conf, custom_objects=custom_objects)
            for conf in config.values()
        ]

        return SequentialBlock(layers)

    def __rrshift__(self, other):
        return right_shift_layer(self, other)

    def __rshift__(self, other):
        # pylint: disable=arguments-out-of-order
        return right_shift_layer(other, self)


@tf.keras.utils.register_keras_serializable(package="merlin.models")
class ParallelBlock(TabularBlock):
    """Merge multiple layers or TabularModule's into a single output of TabularData.

    Parameters
    ----------
    blocks_to_merge: Union[TabularModule, Dict[str, TabularBlock]]
        TabularBlocks to merge into, this can also be one or multiple dictionaries keyed by the
        name the module should have.
    {tabular_module_parameters}
    """

    def __init__(
        self,
        *inputs: Union[tf.keras.layers.Layer, Dict[str, tf.keras.layers.Layer]],
        pre: Optional[BlockType] = None,
        post: Optional[BlockType] = None,
        aggregation: Optional[TabularAggregationType] = None,
        schema: Optional[Schema] = None,
        name: Optional[str] = None,
        strict: bool = False,
        automatic_pruning: bool = True,
        **kwargs,
    ):
        super().__init__(
            pre=pre, post=post, aggregation=aggregation, schema=schema, name=name, **kwargs
        )
        self.strict = strict
        self.automatic_pruning = automatic_pruning
        self.parallel_layers: Union[List[TabularBlock], Dict[str, TabularBlock]]
        if isinstance(inputs, tuple) and len(inputs) == 1 and isinstance(inputs[0], (list, tuple)):
            inputs = inputs[0]
        if all(isinstance(x, dict) for x in inputs):
            to_merge: Dict[str, tf.keras.layers.Layer] = reduce(
                lambda a, b: dict(a, **b), inputs
            )  # type: ignore
            parsed_to_merge: Dict[str, TabularBlock] = {}
            for key, val in to_merge.items():
                parsed_to_merge[key] = val
            self.parallel_layers = parsed_to_merge
        elif all(isinstance(x, tf.keras.layers.Layer) for x in inputs):
            parsed: List[TabularBlock] = []
            for i, inp in enumerate(inputs):
                parsed.append(inp)  # type: ignore
            self.parallel_layers = parsed
        else:
            raise ValueError(
                "Please provide one or multiple layer's to merge or "
                f"dictionaries of layer. got: {inputs}"
            )

        # Merge schemas if necessary.
        if not schema and all(getattr(m, "_schema", False) for m in self.parallel_values):
            if len(self.parallel_values) == 1:
                self.set_schema(self.parallel_values[0].schema)
            else:
                s = reduce(
                    lambda a, b: a + b, [m.schema for m in self.parallel_values]
                )  # type: ignore
                self.set_schema(s)

    @property
    def parallel_values(self) -> List[tf.keras.layers.Layer]:
        if isinstance(self.parallel_layers, dict):
            return list(self.parallel_layers.values())

        return self.parallel_layers

    @property
    def parallel_dict(self) -> Dict[Union[str, int], tf.keras.layers.Layer]:
        if isinstance(self.parallel_layers, dict):
            return self.parallel_layers

        return {i: m for i, m in enumerate(self.parallel_layers)}

    @property
    def layers(self) -> List[tf.keras.layers.Layer]:
        return self.parallel_values

    def select_by_name(self, name: str) -> Optional["Block"]:
        """Select a parallel block by name

        Returns
        -------
        Block
            The block corresponding to the name
        """
        return self.parallel_dict.get(name)

    def __getitem__(self, key) -> "Block":
        return self.parallel_dict[key]

    def __setitem__(self, key: str, item: "Block"):
        self.parallel_dict[key] = item

    def add_branch(self, name: str, block: "Block") -> "ParallelBlock":
        if isinstance(self.parallel_layers, dict):
            self.parallel_layers[name] = block

        return self

    def apply_to_branch(self, branch_name: str, *block: "Block"):
        if isinstance(self.parallel_layers, dict):
            self.parallel_layers[branch_name] = self.parallel_layers[branch_name].apply(*block)

    def call(self, inputs, **kwargs):
        """The call method for ParallelBlock

        Parameters
        ----------
        inputs : TabularData
            The inputs for the Parallel Block

        Returns
        -------
        TabularData
            Outputs of the ParallelBlock
        """
        if self.strict:
            assert isinstance(inputs, dict), "Inputs needs to be a dict"

        outputs = {}
        if isinstance(inputs, dict) and all(
            name in inputs for name in list(self.parallel_dict.keys())
        ):
            for name, block in self.parallel_dict.items():
                out = call_layer(block, inputs[name], **kwargs)
                if not isinstance(out, dict):
                    out = {name: out}
                outputs.update(out)
        else:
            for name, layer in self.parallel_dict.items():
                out = call_layer(layer, inputs, **kwargs)
                if not isinstance(out, dict):
                    out = {name: out}
                outputs.update(out)

        return outputs

    def compute_call_output_shape(self, input_shape):
        output_shapes = {}

        for name, layer in self.parallel_dict.items():
            if isinstance(input_shape, dict) and all(
                key in input_shape for key in list(self.parallel_dict.keys())
            ):
                out = layer.compute_output_shape(input_shape[name])
            else:
                out = layer.compute_output_shape(input_shape)
            if isinstance(out, dict):
                output_shapes.update(out)
            else:
                output_shapes[name] = out

        return output_shapes

    def build(self, input_shape):
        to_prune = []
        propagate_all_inputs = True
        if isinstance(input_shape, dict) and all(
            name in input_shape for name in list(self.parallel_dict.keys())
        ):
            propagate_all_inputs = False

        for key, branch in self.parallel_dict.items():
            branch_shape = input_shape[key] if not propagate_all_inputs else input_shape
            branch.build(branch_shape)
            branch_out_shape = branch.compute_output_shape(branch_shape)
            if self.automatic_pruning and branch_out_shape == {}:
                to_prune.append(key)

        for p in to_prune:
            self.parallel_layers.pop(p)

        return super().build(input_shape)

    def get_config(self):
        config = super(ParallelBlock, self).get_config()
        config.update({"automatic_pruning": self.automatic_pruning})

        return tf_utils.maybe_serialize_keras_objects(self, config, ["parallel_layers"])

    @classmethod
    def parse_config(cls, config, custom_objects=None):
        config = tf_utils.maybe_deserialize_keras_objects(config, ["pre", "post", "aggregation"])
        if "schema" in config:
            config["schema"] = schema_utils.tensorflow_metadata_json_to_schema(config["schema"])

        parallel_layers = config.pop("parallel_layers")
        if isinstance(parallel_layers, dict):
            inputs = {
                name: tf.keras.layers.deserialize(conf, custom_objects=custom_objects)
                for name, conf in parallel_layers.items()
            }
        elif isinstance(parallel_layers, (list, tuple)):
            inputs = [
                tf.keras.layers.deserialize(conf, custom_objects=custom_objects)
                for conf in parallel_layers
            ]
        else:
            raise ValueError("Parallel layers need to be a list or a dict")

        return inputs, config

    @classmethod
    def from_config(cls, config, custom_objects=None):
        inputs, config = cls.parse_config(config, custom_objects)

        return cls(inputs, **config)


@tf.keras.utils.register_keras_serializable(package="merlin.models")
class WithShortcut(ParallelBlock):
    def __init__(
        self,
        block: Union[tf.keras.layers.Layer, Block],
        shortcut_filter: Optional[Filter] = None,
        aggregation=None,
        post: Optional[BlockType] = None,
        schema: Optional[Schema] = None,
        name: Optional[str] = None,
        strict: bool = False,
        block_outputs_name: Optional[str] = None,
        **kwargs,
    ):
        block_outputs_name = block_outputs_name or block.name
        shortcut = shortcut_filter if shortcut_filter else NoOp()
        inputs = {block_outputs_name: block, "shortcut": shortcut}
        super().__init__(
            inputs,
            post=post,
            aggregation=aggregation,
            schema=schema,
            name=name,
            strict=strict,
            **kwargs,
        )

    @classmethod
    def from_config(cls, config, **kwargs):
        output = ParallelBlock.from_config(config, **kwargs)
        output.__class__ = cls

        return output


@tf.keras.utils.register_keras_serializable(package="merlin.models")
class ResidualBlock(WithShortcut):
    def __init__(
        self,
        block: Union[tf.keras.layers.Layer, Block],
        activation=None,
        post: Optional[BlockType] = None,
        schema: Optional[Schema] = None,
        name: Optional[str] = None,
        strict: bool = False,
        **kwargs,
    ):
        from merlin.models.tf.blocks.core.aggregation import SumResidual

        super().__init__(
            block,
            post=post,
            aggregation=SumResidual(activation=activation),
            schema=schema,
            name=name,
            strict=strict,
            **kwargs,
        )



import tempfile
import typing as tp

import numpy as np
import tensorflow as tf

from merlin.core.dispatch import DataFrameType, concat_columns, get_lib
from merlin.models.tf.blocks.core.base import Block
from merlin.models.tf.dataset import BatchedDataset
from merlin.models.tf.models.base import Model, RetrievalModel
from merlin.models.utils.schema_utils import select_targets
from merlin.schema import Schema, Tags


class ModelEncode:
    def __init__(
        self,
        model,
        output_names=None,
        data_iterator_func=None,
        model_load_func=None,
        model_encode_func=None,
        output_concat_func=None,
    ):
        super().__init__()
        self._model = model
        self.output_names = [output_names] if isinstance(output_names, str) else output_names
        self.data_iterator_func = data_iterator_func
        self.model_load_func = model_load_func
        self.model_encode_func = model_encode_func
        self.output_concat_func = output_concat_func

    @property
    def model(self):
        if isinstance(self._model, str):
            self._model = self.model_load_func(self._model)
        return self._model

    def __call__(
        self,
        df: DataFrameType,
        filter_input_columns: tp.Optional[tp.List[str]] = None,
        filter_output_columns: tp.Optional[tp.List[str]] = None,
    ) -> DataFrameType:
        # Set defaults
        iterator_func = self.data_iterator_func or (lambda x: [x])
        encode_func = self.model_encode_func or (lambda x, y: x(y))
        concat_func = self.output_concat_func or np.concatenate

        # Iterate over batches of df and collect predictions
        outputs = concat_func([encode_func(self.model, batch) for batch in iterator_func(df)])
        output_names = self.output_names or [str(i) for i in range(outputs.shape[1])]
        model_output_df = type(df)(outputs, columns=output_names)
        if filter_output_columns:
            model_output_df = model_output_df[filter_output_columns]
        input_df = df if not filter_input_columns else df[filter_input_columns]

        output_df = concat_columns([input_df, model_output_df])

        return output_df

    def transform(self, col_selector, df: DataFrameType, **kwargs) -> DataFrameType:
        return self(df[col_selector], **kwargs)


class TFModelEncode(ModelEncode):
    def __init__(
        self,
        model: tp.Union[Model, tf.keras.Model],
        output_names: tp.Optional[tp.List[str]] = None,
        batch_size: int = 512,
        save_path: tp.Optional[str] = None,
        block_load_func: tp.Optional[tp.Callable[[str], Block]] = None,
        schema: tp.Optional[Schema] = None,
        output_concat_func=None,
    ):
        save_path = save_path or tempfile.mkdtemp()
        model.save(save_path)

        model_load_func = block_load_func if block_load_func else tf.keras.models.load_model
        if not output_names:
            try:
                output_names = model.block.last.task_names
            except AttributeError:
                pass
        if not output_concat_func:
            if len(output_names) == 1:  # type: ignore
                output_concat_func = np.concatenate
            else:
                output_concat_func = get_lib().concat  # type: ignore

        self.schema = schema or model.schema

        super().__init__(
            save_path,
            output_names,
            data_iterator_func=data_iterator_func(self.schema, batch_size=batch_size),
            model_load_func=model_load_func,
            model_encode_func=model_encode,
            output_concat_func=output_concat_func,
        )

    # def fit_transform(self, data) -> nvt.Dataset:
    #     features = self.schema.column_names >> self
    #
    #     # Fit and transform
    #     processor = nvt.Workflow(features)
    #     output = processor.fit_transform(data)
    #
    #     return output


class ItemEmbeddings(TFModelEncode):
    def __init__(self, model: Model, batch_size: int = 512, save_path: tp.Optional[str] = None):
        item_block = model.block.first.item_block()
        schema = item_block.schema

        super().__init__(
            item_block,
            save_path=save_path,
            batch_size=batch_size,
            schema=schema,
            output_concat_func=np.concatenate,
        )


class QueryEmbeddings(TFModelEncode):
    def __init__(
        self,
        model: RetrievalModel,
        batch_size: int = 512,
        save_path: tp.Optional[str] = None,
    ):
        query_block = model.block.first.query_block()
        schema = query_block.schema

        super().__init__(
            query_block,
            save_path=save_path,
            batch_size=batch_size,
            schema=schema,
            output_concat_func=np.concatenate,
        )


def model_encode(model, batch):
    # TODO: How to handle list outputs?

    model_outputs = model(batch[0])

    if isinstance(model_outputs, dict):
        return get_lib().DataFrame({key: encode_output(val) for key, val in model_outputs.items()})

    return encode_output(model_outputs)


def encode_output(output: tf.Tensor):
    if len(output.shape) == 2 and output.shape[1] == 1:
        output = tf.squeeze(output)

    return output.numpy()


def data_iterator_func(schema, batch_size: int = 512):
    import merlin.io.dataset

    cat_cols = schema.select_by_tag(Tags.CATEGORICAL).excluding_by_tag(Tags.TARGET).column_names
    cont_cols = schema.select_by_tag(Tags.CONTINUOUS).excluding_by_tag(Tags.TARGET).column_names
    targets = select_targets(schema).column_names

    def data_iterator(dataset):
        return BatchedDataset(
            merlin.io.dataset.Dataset(dataset),
            batch_size=batch_size,
            cat_names=cat_cols,
            cont_names=cont_cols,
            label_names=targets,
            shuffle=False,
        )

    return data_iterator



#
# Copyright (c) 2021, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import numpy as np
import pytest
from tensorflow.keras.initializers import RandomUniform

import merlin.models.tf as mm
from merlin.io import Dataset
from merlin.models.tf.utils import testing_utils
from merlin.schema import Tags


def test_embedding_features(tf_cat_features):
    dim = 15
    feature_config = {
        f: mm.FeatureConfig(mm.TableConfig(100, dim, name=f, initializer=None))
        for f in tf_cat_features.keys()
    }
    embeddings = mm.EmbeddingFeatures(feature_config)(tf_cat_features)

    assert list(embeddings.keys()) == list(feature_config.keys())
    assert all([emb.shape[-1] == dim for emb in embeddings.values()])


def test_embedding_features_yoochoose(testing_data: Dataset):
    schema = testing_data.schema.select_by_tag(Tags.CATEGORICAL)

    emb_module = mm.EmbeddingFeatures.from_schema(
        schema,
        embedding_options=mm.EmbeddingOptions(embedding_dim_default=512),
    )
    embeddings = emb_module(mm.sample_batch(testing_data, batch_size=100, include_targets=False))

    assert sorted(list(embeddings.keys())) == sorted(schema.column_names)
    assert all(emb.shape[-1] == 512 for emb in embeddings.values())
    max_value = list(schema.select_by_name("item_id"))[0].int_domain.max
    assert emb_module.embedding_tables["item_id"].shape[0] == max_value + 1

    # These embeddings have not a specific initializer, so they should
    # have default truncated normal initialization
    default_truncated_normal_std = 0.05
    for emb_key in embeddings:
        assert embeddings[emb_key].numpy().mean() == pytest.approx(0.0, abs=0.02)
        assert embeddings[emb_key].numpy().std() == pytest.approx(
            default_truncated_normal_std, abs=0.04
        )


def test_serialization_embedding_features(testing_data: Dataset):
    inputs = mm.EmbeddingFeatures.from_schema(testing_data.schema)

    copy_layer = testing_utils.assert_serialization(inputs)

    assert list(inputs.feature_config.keys()) == list(copy_layer.feature_config.keys())

    from merlin.models.tf.inputs.embedding import serialize_table_config as serialize

    assert all(
        serialize(inputs.feature_config[key].table)
        == serialize(copy_layer.feature_config[key].table)
        for key in copy_layer.feature_config
    )


@testing_utils.mark_run_eagerly_modes
def test_embedding_features_yoochoose_model(music_streaming_data: Dataset, run_eagerly):
    schema = music_streaming_data.schema.select_by_tag(Tags.CATEGORICAL)

    inputs = mm.EmbeddingFeatures.from_schema(schema, aggregation="concat")
    body = mm.SequentialBlock([inputs, mm.MLPBlock([64])])
    model = mm.Model(body, mm.BinaryClassificationTask("click"))

    testing_utils.model_test(model, music_streaming_data, run_eagerly=run_eagerly)


def test_embedding_features_yoochoose_custom_dims(testing_data: Dataset):
    schema = testing_data.schema.select_by_tag(Tags.CATEGORICAL)

    emb_module = mm.EmbeddingFeatures.from_schema(
        schema,
        embedding_options=mm.EmbeddingOptions(
            embedding_dims={"item_id": 100}, embedding_dim_default=64
        ),
    )

    embeddings = emb_module(mm.sample_batch(testing_data, batch_size=100, include_targets=False))

    assert len(emb_module.losses) == 0, "There should be no regularization loss by default"

    assert emb_module.embedding_tables["item_id"].shape[1] == 100
    assert emb_module.embedding_tables["categories"].shape[1] == 64

    assert embeddings["item_id"].shape[1] == 100
    assert embeddings["categories"].shape[1] == 64


def test_embedding_features_l2_reg(testing_data: Dataset):
    schema = testing_data.schema.select_by_tag(Tags.CATEGORICAL)

    emb_module = mm.EmbeddingFeatures.from_schema(
        schema,
        embedding_options=mm.EmbeddingOptions(
            embedding_dims={"item_id": 100}, embedding_dim_default=64, embeddings_l2_reg=0.1
        ),
    )

    _ = emb_module(mm.sample_batch(testing_data, batch_size=100, include_targets=False))

    l2_emb_losses = emb_module.losses

    assert len(l2_emb_losses) == len(
        schema
    ), "The number of reg losses should equal to the number of embeddings"

    for reg_loss in l2_emb_losses:
        assert reg_loss > 0.0


def test_embedding_features_yoochoose_infer_embedding_sizes(testing_data: Dataset):
    schema = testing_data.schema.select_by_tag(Tags.CATEGORICAL)

    emb_module = mm.EmbeddingFeatures.from_schema(
        schema,
        embedding_options=mm.EmbeddingOptions(
            infer_embedding_sizes=True, infer_embedding_sizes_multiplier=3.0
        ),
    )

    embeddings = emb_module(mm.sample_batch(testing_data, batch_size=100, include_targets=False))

    assert emb_module.embedding_tables["user_id"].shape[1] == embeddings["user_id"].shape[1] == 20
    assert (
        emb_module.embedding_tables["user_country"].shape[1]
        == embeddings["user_country"].shape[1]
        == 9
    )
    assert emb_module.embedding_tables["item_id"].shape[1] == embeddings["item_id"].shape[1] == 46
    assert (
        emb_module.embedding_tables["categories"].shape[1]
        == embeddings["categories"].shape[1]
        == 13
    )


def test_embedding_features_yoochoose_infer_embedding_sizes_multiple_8(testing_data: Dataset):
    schema = testing_data.schema.select_by_tag(Tags.CATEGORICAL)

    emb_module = mm.EmbeddingFeatures.from_schema(
        schema,
        embedding_options=mm.EmbeddingOptions(
            infer_embedding_sizes=True,
            infer_embedding_sizes_multiplier=3.0,
            infer_embeddings_ensure_dim_multiple_of_8=True,
        ),
    )

    embeddings = emb_module(mm.sample_batch(testing_data, batch_size=100, include_targets=False))

    assert emb_module.embedding_tables["user_id"].shape[1] == embeddings["user_id"].shape[1] == 24
    assert (
        emb_module.embedding_tables["user_country"].shape[1]
        == embeddings["user_country"].shape[1]
        == 16
    )
    assert emb_module.embedding_tables["item_id"].shape[1] == embeddings["item_id"].shape[1] == 48
    assert (
        emb_module.embedding_tables["categories"].shape[1]
        == embeddings["categories"].shape[1]
        == 16
    )


def test_embedding_features_yoochoose_partially_infer_embedding_sizes(testing_data: Dataset):
    schema = testing_data.schema.select_by_tag(Tags.CATEGORICAL)

    emb_module = mm.EmbeddingFeatures.from_schema(
        schema,
        embedding_options=mm.EmbeddingOptions(
            embedding_dims={"user_id": 50, "user_country": 100},
            infer_embedding_sizes=True,
            infer_embedding_sizes_multiplier=3.0,
        ),
    )

    embeddings = emb_module(mm.sample_batch(testing_data, batch_size=100, include_targets=False))

    assert emb_module.embedding_tables["user_id"].shape[1] == embeddings["user_id"].shape[1] == 50
    assert (
        emb_module.embedding_tables["user_country"].shape[1]
        == embeddings["user_country"].shape[1]
        == 100
    )
    assert emb_module.embedding_tables["item_id"].shape[1] == embeddings["item_id"].shape[1] == 46
    assert (
        emb_module.embedding_tables["categories"].shape[1]
        == embeddings["categories"].shape[1]
        == 13
    )


def test_embedding_features_yoochoose_custom_initializers(testing_data: Dataset):
    schema = testing_data.schema.select_by_tag(Tags.CATEGORICAL)

    random_max_abs_value = 0.3
    emb_module = mm.EmbeddingFeatures.from_schema(
        schema,
        embedding_options=mm.EmbeddingOptions(
            embedding_dim_default=512,
            embeddings_initializers={
                "user_id": RandomUniform(minval=-random_max_abs_value, maxval=random_max_abs_value),
                "user_country": RandomUniform(
                    minval=-random_max_abs_value, maxval=random_max_abs_value
                ),
            },
        ),
    )

    embeddings = emb_module(mm.sample_batch(testing_data, batch_size=100, include_targets=False))

    assert embeddings["user_id"].numpy().min() == pytest.approx(-random_max_abs_value, abs=0.02)
    assert embeddings["user_id"].numpy().max() == pytest.approx(random_max_abs_value, abs=0.02)

    assert embeddings["user_country"].numpy().min() == pytest.approx(
        -random_max_abs_value, abs=0.02
    )
    assert embeddings["user_country"].numpy().max() == pytest.approx(random_max_abs_value, abs=0.02)

    # These embeddings have not a specific initializer, so they should
    # have default truncated normal initialization
    default_truncated_normal_std = 0.05
    assert embeddings["item_id"].numpy().mean() == pytest.approx(0.0, abs=0.02)
    assert embeddings["item_id"].numpy().std() == pytest.approx(
        default_truncated_normal_std, abs=0.04
    )

    assert embeddings["categories"].numpy().mean() == pytest.approx(0.0, abs=0.02)
    assert embeddings["categories"].numpy().std() == pytest.approx(
        default_truncated_normal_std, abs=0.04
    )


def test_embedding_features_yoochoose_pretrained_initializer(testing_data: Dataset):
    schema = testing_data.schema.select_by_tag(Tags.CATEGORICAL)

    pretrained_emb_item_ids = np.random.random((51997, 64))
    pretrained_emb_categories = np.random.random((332, 64))

    emb_module = mm.EmbeddingFeatures.from_schema(
        schema,
        embedding_options=mm.EmbeddingOptions(
            embeddings_initializers={
                "item_id": mm.TensorInitializer(pretrained_emb_item_ids),
                "categories": mm.TensorInitializer(pretrained_emb_categories),
            },
        ),
    )

    # Calling the first batch, so that embedding tables are build
    _ = emb_module(mm.sample_batch(testing_data, batch_size=10, include_targets=False))

    assert np.allclose(emb_module.embedding_tables["item_id"].numpy(), pretrained_emb_item_ids)
    assert np.allclose(emb_module.embedding_tables["categories"].numpy(), pretrained_emb_categories)


def test_embedding_features_exporting_and_loading_pretrained_initializer(testing_data: Dataset):
    schema = testing_data.schema.select_by_tag(Tags.CATEGORICAL)
    emb_module = mm.EmbeddingFeatures.from_schema(schema)

    # Calling the first batch, so that embedding tables are build
    _ = emb_module(mm.sample_batch(testing_data, batch_size=10, include_targets=False))
    item_id_embeddings = emb_module.embedding_tables["item_id"]

    items_embeddings_dataset = emb_module.embedding_table_dataset(Tags.ITEM_ID, gpu=False)
    assert np.allclose(
        item_id_embeddings.numpy(), items_embeddings_dataset.to_ddf().compute().values
    )

    emb_init = mm.TensorInitializer.from_dataset(items_embeddings_dataset)
    assert np.allclose(item_id_embeddings.numpy(), emb_init(item_id_embeddings.shape).numpy())

    # Test GPU export if available
    try:
        import cudf  # noqa: F401

        items_embeddings_dataset = emb_module.embedding_table_dataset(Tags.ITEM_ID, gpu=True)
        assert np.allclose(
            item_id_embeddings.numpy(),
            items_embeddings_dataset.to_ddf().compute().to_pandas().values,
        )

        emb_init = mm.TensorInitializer.from_dataset(items_embeddings_dataset)
        assert np.allclose(item_id_embeddings.numpy(), emb_init(item_id_embeddings.shape).numpy())

    except ImportError:
        pass


def test_shared_embeddings(music_streaming_data: Dataset):
    inputs = mm.InputBlock(music_streaming_data.schema)

    embeddings = inputs.select_by_name(Tags.CATEGORICAL.value)

    assert embeddings.table_config("item_genres") == embeddings.table_config("user_genres")



#
# Copyright (c) 2021, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from typing import Dict, Optional

import tensorflow as tf
from tensorflow.python.ops import embedding_ops

from merlin.models.tf.blocks.sampling.base import EmbeddingWithMetadata, ItemSampler
from merlin.models.tf.blocks.sampling.queue import FIFOQueue
from merlin.models.tf.typing import TabularData


class CachedCrossBatchSampler(ItemSampler):
    """Provides efficient cached cross-batch [1]_ / inter-batch [2]_ negative sampling
    for two-tower item retrieval model. The caches consists of a fixed capacity FIFO queue
    which keeps the item embeddings from the last N batches. All items in the queue are
    sampled as negatives for upcoming batches.
    It is more efficient than computing embeddings exclusively for negative items.
    This is a popularity-biased sampling as popular items are observed more often
    in training batches.
    Compared to `InBatchSampler`, the `CachedCrossBatchSampler` allows for larger number
    of negative items, not limited to the batch size. The gradients are not computed
    for the cached negative embeddings which is a scalable approach. A common combination
    of samplers for the `ItemRetrievalScorer` is `[InBatchSampler(),
    CachedCrossBatchSampler(ignore_last_batch_on_sample=True)]`, which computes gradients for
    the in-batch negatives and not for the cached item embeddings.
    P.s. Ignoring the false negatives (negative items equal to the positive ones) is
    managed by `ItemRetrievalScorer(..., sampling_downscore_false_negatives=True)`

    References
    ----------
    .. [1] Wang, Jinpeng, Jieming Zhu, and Xiuqiang He. "Cross-Batch Negative Sampling
       for Training Two-Tower Recommenders." Proceedings of the 44th International ACM
       SIGIR Conference on Research and Development in Information Retrieval. 2021.

    .. [2] Zhou, Chang, et al. "Contrastive learning for debiased candidate generation
       in large-scale recommender systems." Proceedings of the 27th ACM SIGKDD Conference
       on Knowledge Discovery & Data Mining. 2021.

    Parameters
    ----------
    capacity: int
        The queue capacity to store samples
    ignore_last_batch_on_sample: bool
        Whether should include the last batch in the sampling. By default `False`,
        as for sampling from the current batch we recommend `InBatchSampler()`, which
        allows computing gradients for in-batch negative items
    """

    def __init__(
        self,
        capacity: int,
        ignore_last_batch_on_sample: bool = True,
        **kwargs,
    ):
        assert capacity > 0
        super().__init__(max_num_samples=capacity, **kwargs)
        self.ignore_last_batch_on_sample = ignore_last_batch_on_sample
        self.item_metadata_dtypes: Dict[str, tf.dtypes.DType] = {}

        self._last_batch_size = 0
        self._item_embeddings_queue: Optional[FIFOQueue] = None

    @property
    def item_embeddings_queue(self) -> FIFOQueue:
        if not self._item_embeddings_queue:
            raise ValueError("Item embeddings queue is not initialized")

        return self._item_embeddings_queue

    def _maybe_build(self, inputs: TabularData) -> None:
        items_metadata = inputs["metadata"]
        if not self.item_metadata_dtypes:
            self.item_metadata_dtypes = {
                feat_name: items_metadata[feat_name].dtype for feat_name in items_metadata
            }
        super()._maybe_build(inputs)

    def build(self, input_shapes: TabularData) -> None:
        queue_size = self.max_num_samples

        item_embeddings_dims = list(input_shapes["embeddings"][1:])
        self._item_embeddings_queue = FIFOQueue(
            capacity=queue_size,
            dims=item_embeddings_dims,
            dtype=tf.float32,
            name="item_emb",
        )

        self.items_metadata_queue = dict()
        items_metadata = input_shapes["metadata"]
        for feat_name in items_metadata:
            dims = []
            if items_metadata[feat_name].dims is not None:
                dims = list(items_metadata[feat_name][1:])
            self.items_metadata_queue[feat_name] = FIFOQueue(
                capacity=queue_size,
                dims=dims,
                dtype=self.item_metadata_dtypes[feat_name],
                name=f"item_metadata_{feat_name}",
            )

    def _check_built(self) -> None:
        if self._item_embeddings_queue is None:
            raise Exception(
                f"The {self.__class__.__name__} layer was not built yet. "
                "You need to call() that layer at least once before "
                "so that it is built before calling add() or sample() directly"
            )

    def call(self, inputs: TabularData, training=True) -> EmbeddingWithMetadata:
        """Adds the current batch to the FIFO queue cache and samples all items
        embeddings from the last N cached batches.

        Parameters
        ----------
        inputs : TabularData
            Dict with two keys:
              "items_embeddings": Items embeddings tensor
              "items_metadata": Dict like `{"<feature name>": "<feature tensor>"}` which contains
              features that might be relevant for the sampler (e.g. item id, item popularity, item
              recency).
              The `CachedCrossBatchSampler` does not use metadata features
              specifically, but "item_id" is required when using in combination with
              `ItemRetrievalScorer(..., sampling_downscore_false_negatives=True)`, so that
              false negatives are identified and downscored.
        training : bool, optional
            Flag indicating if on training mode, by default True

        Returns
        -------
        EmbeddingWithMetadata
            Value object with the sampled item embeddings and item metadata
        """
        if self.ignore_last_batch_on_sample:
            items_embeddings = self.sample()
            self.add(inputs, training)
        else:
            self.add(inputs, training)
            items_embeddings = self.sample()

        return items_embeddings

    def add(  # type: ignore
        self,
        inputs: TabularData,
        training: bool = True,
    ) -> None:
        self._check_built()

        if training:
            self._check_inputs_batch_sizes(inputs)
            items_embeddings = inputs["embeddings"]
            items_metadata = inputs["metadata"]

            self.item_embeddings_queue.enqueue_many(items_embeddings)
            for feat_name in items_metadata:
                self.items_metadata_queue[feat_name].enqueue_many(items_metadata[feat_name])

            self._last_batch_size = tf.shape(items_embeddings)[0]

    def sample(self) -> EmbeddingWithMetadata:
        self._check_built()
        items_embeddings = self.item_embeddings_queue.list_all()
        items_metadata = {
            feat_name: self.items_metadata_queue[feat_name].list_all()
            for feat_name in self.items_metadata_queue
        }

        return EmbeddingWithMetadata(
            items_embeddings,
            items_metadata,
        )


class CachedUniformSampler(CachedCrossBatchSampler):
    """Provides a cached uniform negative sampling for two-tower item retrieval model.
    It is similar to the `CachedCrossBatchSampler`, the main difference
    is that the `CachedUniformSampler` is a popularity-based sampler and `CachedUniformSampler`
    only keeps unique item embeddings in the queue for uniform sampling.
    The caches consists of two fixed capacity internal FIFO queues that hold both
    item ids and item embeddings.
    It ensures that each item id (and corresponding embedding) is added only once into
    the queue. If the item id was already included in the queue by a previous batch,
    the embedding is updated.
    As the queues reach their capacity of unique items, new items will replace the
    first items added to the queue.

    This is a cached implementation of [1]_, where those authors proposed combining
    in-batch sampling (our `InBatchSampler()`) with uniform sampling. Differently from
    [1] which requires a separate dataset with the all unique items (and corresponding features)
    to generate the item embeddings, our streaming approach in `CachedUniformSampler` keeps
    caching new items as they appear in the batches. That means that the very first
    processed batches will have less negative samples.

    P.s. Ignoring the false negatives (negative items equal to the positive ones) is
    managed by `ItemRetrievalScorer(..., sampling_downscore_false_negatives=True)`

    References
    ----------
    .. [1] Yang, Ji, et al. "Mixed negative sampling for learning two-tower neural networks in
       recommendations." Companion Proceedings of the Web Conference 2020. 2020.

    Parameters
    ----------
    capacity: int
        The queue capacity to store samples
    ignore_last_batch_on_sample: bool
        Whether should include the last batch in the sampling. By default `False`,
        as for sampling from the current batch we recommend `InBatchSampler()`, which
        allows computing gradients for in-batch negative items
    item_id_feature_name: str
        Name of the column containing the item ids
        Defaults to `item_id`
    """

    def __init__(
        self,
        capacity: int,
        ignore_last_batch_on_sample: bool = True,
        item_id_feature_name: str = "item_id",
        **kwargs,
    ):
        super().__init__(
            capacity=capacity, ignore_last_batch_on_sample=ignore_last_batch_on_sample, **kwargs
        )
        self.item_id_feature_name = item_id_feature_name

    def _check_inputs(self, inputs):
        assert (
            str(self.item_id_feature_name) in inputs["metadata"]
        ), "The 'item_id' metadata feature is required by UniformSampler."

    def add(  # type: ignore
        self,
        inputs: TabularData,
        training: bool = True,
    ) -> None:
        """Updates the FIFO queue with batch item embeddings (for items whose ids were
        already added to the queue) and adds to the queue the items seen for the first time

        Parameters
        ----------
        inputs : TabularData
            Dict with two keys:
              "items_embeddings": Items embeddings tensor
              "items_metadata": Dict like `{"<feature name>": "<feature tensor>"}` which contains
              features that might be relevant for the sampler (e.g. item id, item popularity, item
              recency).
              The `CachedUniformSampler` requires the "item_id" feature to identify items already
              added to the queue.
        training : bool, optional
            Flag indicating if on training mode, by default True
        """

        self._check_built()

        if training:
            self._check_inputs(inputs)

            # Removing from inputs repeated item ids and corresponding embeddings
            unique_items = self._get_unique_items(inputs["embeddings"], inputs["metadata"])
            unique_items_ids = unique_items.metadata[self.item_id_feature_name]

            # Identifying what are the internal indices of the item ids in its queue
            item_ids_idxs = self.items_metadata_queue[self.item_id_feature_name].index_of(
                unique_items_ids
            )
            # Creating masks for new and existing items
            new_items_mask = tf.equal(item_ids_idxs, -1)
            existing_items_mask = tf.logical_not(new_items_mask)

            update_indices = tf.expand_dims(item_ids_idxs[existing_items_mask], -1)
            # Updating embeddings of existing items
            self.item_embeddings_queue.update_by_indices(
                indices=update_indices,
                values=unique_items.embeddings[existing_items_mask],
            )

            for feat_name in self.items_metadata_queue:
                self.items_metadata_queue[feat_name].update_by_indices(
                    indices=update_indices,
                    values=unique_items.metadata[feat_name][existing_items_mask],
                )

            # Adding to the queue items not found
            new_items = EmbeddingWithMetadata(
                embeddings=unique_items.embeddings[new_items_mask],
                metadata={
                    feat_name: unique_items.metadata[feat_name][new_items_mask]
                    for feat_name in unique_items.metadata
                },
            )
            super().add(new_items.__dict__, training)

    def _get_unique_items(
        self, items_embeddings: tf.Tensor, items_metadata: TabularData
    ) -> EmbeddingWithMetadata:
        """Extracts the embeddings and corresponding metadata features for
        the unique items found in the batch, based on the item ids

        Parameters
        ----------
        items_embeddings : items_metadata
            Batch item embeddings
        items_metadata : TabularData
            Batch item metadata features

        Returns
        -------
        EmbeddingWithMetadata
            An EmbeddingWithMetadata object with the embeddings and
            metadata features of the unique items
        """
        item_ids = items_metadata[self.item_id_feature_name]
        sorting_item_ids_idx = tf.argsort(item_ids)

        sorted_embeddings = tf.gather(items_embeddings, sorting_item_ids_idx)
        sorted_metadata = {
            feat_name: tf.gather(items_metadata[feat_name], sorting_item_ids_idx)
            for feat_name in items_metadata
        }

        sorted_item_ids = tf.gather(item_ids, sorting_item_ids_idx)

        non_repeated_item_ids_mask = tf.concat(
            [
                # First element is never repeated
                tf.constant(
                    True,
                    shape=(1,),
                ),
                tf.not_equal(sorted_item_ids[:-1], sorted_item_ids[1:]),
            ],
            axis=0,
        )

        output_embeddings = sorted_embeddings[non_repeated_item_ids_mask]
        output_metadata = {
            feat_name: items_metadata[feat_name][non_repeated_item_ids_mask]
            for feat_name in sorted_metadata
        }

        return EmbeddingWithMetadata(
            embeddings=output_embeddings,
            metadata=output_metadata,
        )


class PopularityBasedSampler(ItemSampler):
    """
    Provides a popularity-based negative sampling for the softmax layer
    to ensure training efficiency when the catalog of items is very large.
    The capacity of the queue is fixed and is equal to the catalog size.
    For each batch, we sample `max_num_samples` unique negatives.

    This implementation does not require the actual item frequencies/probabilities
    distribution, but instead tries to approximate the item
    probabilities using the log_uniform (zipfian) distribution.
    The only requirement is that the item ids are decreasingly sorted by their count frequency.
    We use the default log-uniform (zipfian) sampler given by Tensorflow:
        [log_uniform_candidate_sampler](https://www.tensorflow.org/api_docs/python/tf/random/log_uniform_candidate_sampler)
    The `Categorify` op provided by nvtabular supports the frequency-based encoding as default.

    P.s. Ignoring the false negatives (negative items equal to the positive ones) is
    managed by `ItemRetrievalScorer(..., sampling_downscore_false_negatives=True)`

    Parameters
    ----------
    max_num_samples: int
        The number of unique negatives to sample at each batch.
    max_id: int
        The maximum id value to be sampled. It should be equal to the
        categorical feature cardinality
    min_id: int
        The minimum id value to be sampled. Useful to ignore the first categorical
        encoded ids, which are usually reserved for <nulls>, out-of-vocabulary or padding.
        Defaults to 0.
    seed: int
        Fix the random values returned by the sampler to ensure reproducibility
        Defaults to None
    item_id_feature_name: str
        Name of the column containing the item ids
        Defaults to `item_id`
    """

    def __init__(
        self,
        max_id: int,
        min_id: int = 0,
        max_num_samples: int = 100,
        seed: Optional[int] = None,
        item_id_feature_name: str = "item_id",
        **kwargs,
    ):
        super().__init__(max_num_samples=max_num_samples, **kwargs)
        self.max_id = max_id
        self.min_id = min_id
        self.seed = seed
        self.item_id_feature_name = item_id_feature_name

        assert (
            self.max_num_samples <= self.max_id
        ), f"Number of items to sample `{self.max_num_samples}`"
        f" should be less than total number of ids `{self.max_id}`"

    def _check_inputs(self, inputs):
        assert (
            self.item_id_feature_name in inputs["metadata"]
        ), "The 'item_id' metadata feature is required by PopularityBasedSampler."

    def add(self, embeddings: tf.Tensor, items_metadata: TabularData, training=True):
        pass

    def call(
        self, inputs: TabularData, item_weights: tf.Tensor, training=True
    ) -> EmbeddingWithMetadata:
        if training:
            self._check_inputs(inputs)

        tf.assert_equal(
            int(tf.shape(item_weights)[0]),
            self.max_id + 1,
            "The first dimension of the items embeddings "
            f"({int(tf.shape(item_weights)[0])}) and "
            f"the the number of possible classes ({self.max_id+1}) should match.",
        )

        items_embeddings = self.sample(item_weights)
        return items_embeddings

    def _required_features(self):
        return [self.item_id_feature_name]

    def sample(self, item_weights) -> EmbeddingWithMetadata:  # type: ignore
        sampled_ids, _, _ = tf.random.log_uniform_candidate_sampler(
            true_classes=tf.ones((1, 1), dtype=tf.int64),
            num_true=1,
            num_sampled=self.max_num_samples,
            unique=True,
            range_max=self.max_id - self.min_id,
            seed=self.seed,
        )

        # Shifting the sampled ids to ignore the first ids (usually reserved for nulls, OOV)
        sampled_ids += self.min_id

        items_embeddings = embedding_ops.embedding_lookup(item_weights, sampled_ids)

        return EmbeddingWithMetadata(
            items_embeddings,
            metadata={self.item_id_feature_name: tf.cast(sampled_ids, tf.int32)},
        )

    def get_distribution_probs(self):
        """Tries to approximate the log uniform (zipfian) distribution
        used by tf.random.log_uniform_candidate_sampler
        (https://www.tensorflow.org/api_docs/python/tf/random/log_uniform_candidate_sampler)

        Returns
        -------
        tf.Tensor
            A tensor with the expected probability distribution of item ids
            assuming log-uniform (zipfian) distribution
        """
        range_max = self.max_id - self.min_id
        ids = tf.range(0, range_max, dtype=tf.float32)
        estimated_probs = (tf.math.log(ids + 2.0) - tf.math.log(ids + 1.0)) / tf.math.log(
            range_max + 1.0
        )
        # Appending zero(s) in the beginning as padding items should never be samples
        # (thus prob must be zero)
        estimated_probs = tf.concat(
            [tf.zeros(self.min_id + 1, dtype=tf.float32), estimated_probs], axis=0
        )
        return estimated_probs



import pandas as pd
import pytest

import merlin.models.tf as ml
from merlin.io import Dataset


@pytest.mark.parametrize("run_eagerly", [True, False])
def test_model_encode(ecommerce_data: Dataset, run_eagerly):
    body = ml.InputBlock(ecommerce_data.schema).connect(ml.MLPBlock([64]))
    prediction_task = ml.PredictionTasks(ecommerce_data.schema)
    model = ml.Model(body, prediction_task)

    model.compile(run_eagerly=run_eagerly, optimizer="adam")
    model.fit(ecommerce_data, batch_size=50, epochs=1)
    data = model.batch_predict(ecommerce_data, batch_size=10)
    ddf = data.compute(scheduler="synchronous")

    assert len(list(ddf.columns)) == 27
    assert all([task in list(ddf.columns) for task in model.block.last.task_names])


def test_two_tower_embedding_extraction(ecommerce_data: Dataset):
    two_tower = ml.TwoTowerBlock(ecommerce_data.schema, query_tower=ml.MLPBlock([64, 128]))

    model = ml.RetrievalModel(
        two_tower, ml.ItemRetrievalTask(ecommerce_data.schema, target_name="click")
    )
    model.compile(run_eagerly=True, optimizer="adam")
    model.fit(ecommerce_data, batch_size=50, epochs=1)

    item_embs = model.item_embeddings(ecommerce_data, 10)
    item_embs_ddf = item_embs.compute(scheduler="synchronous")

    assert len(list(item_embs_ddf.columns)) == 5 + 128

    user_embs = model.query_embeddings(ecommerce_data, 10)
    user_embs_ddf = user_embs.compute(scheduler="synchronous")

    assert len(list(user_embs_ddf.columns)) == 13 + 128


def test_two_tower_extracted_embeddings_are_equal(ecommerce_data: Dataset):
    import numpy as np

    two_tower = ml.TwoTowerBlock(ecommerce_data.schema, query_tower=ml.MLPBlock([64, 128]))

    model = ml.RetrievalModel(
        two_tower, ml.ItemRetrievalTask(ecommerce_data.schema, target_name="click")
    )
    model.compile(run_eagerly=True, optimizer="adam")
    model.fit(ecommerce_data, batch_size=50, epochs=1)

    item_embs_1 = model.item_embeddings(ecommerce_data, batch_size=10).compute()
    item_embs_2 = model.item_embeddings(ecommerce_data, batch_size=10).compute()

    if not isinstance(item_embs_1, pd.DataFrame):
        item_embs_1 = item_embs_1.to_pandas()
        item_embs_2 = item_embs_2.to_pandas()

    np.testing.assert_array_equal(item_embs_1.values, item_embs_2.values)



#
# Copyright (c) 2021, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import logging
from typing import Dict, Optional, Tuple, Type, Union

from merlin.models.tf.blocks.core.aggregation import SequenceAggregation, SequenceAggregator
from merlin.models.tf.blocks.core.base import Block, BlockType
from merlin.models.tf.blocks.core.combinators import ParallelBlock, TabularAggregationType
from merlin.models.tf.blocks.core.masking import MaskingBlock, masking_registry
from merlin.models.tf.blocks.core.transformations import AsDenseFeatures
from merlin.models.tf.inputs.continuous import ContinuousFeatures
from merlin.models.tf.inputs.embedding import (
    ContinuousEmbedding,
    EmbeddingFeatures,
    EmbeddingOptions,
    SequenceEmbeddingFeatures,
)
from merlin.schema import Schema, Tags, TagsType

LOG = logging.getLogger("merlin-models")


def InputBlock(
    schema: Schema,
    branches: Optional[Dict[str, Block]] = None,
    post: Optional[BlockType] = None,
    aggregation: Optional[TabularAggregationType] = None,
    seq: bool = False,
    max_seq_length: Optional[int] = None,
    add_continuous_branch: bool = True,
    continuous_tags: Optional[Union[TagsType, Tuple[Tags]]] = (Tags.CONTINUOUS,),
    continuous_projection: Optional[Block] = None,
    add_embedding_branch: bool = True,
    embedding_options: EmbeddingOptions = EmbeddingOptions(),
    categorical_tags: Optional[Union[TagsType, Tuple[Tags]]] = (Tags.CATEGORICAL,),
    sequential_tags: Optional[Union[TagsType, Tuple[Tags]]] = (Tags.SEQUENCE,),
    split_sparse: bool = False,
    masking: Optional[Union[str, MaskingBlock]] = None,
    seq_aggregator: Block = SequenceAggregator(SequenceAggregation.MEAN),
    **kwargs,
) -> Block:
    """The entry block of the model to process input features from a schema.

    This function creates continuous and embedding layers, and connects them via `ParallelBlock`.
        If aggregation argument is not set, it returns a dictionary of multiple tensors
        each corresponds to an input feature.
        Otherwise, it merges the tensors into one using the aggregation method.

    Example usage::

        mlp = ml.InputBlock(schema).connect(ml.MLPBlock([64, 32]))

    Parameters:
    ----------
    schema: Schema
        Schema of the input data. This Schema object will be automatically generated using
        [NVTabular](https://nvidia-merlin.github.io/NVTabular/main/Introduction.html).
        Next to this, it's also possible to construct it manually.
    branches: Dict[str, Block], optional
        Dictionary of branches to use inside the InputBlock.
    post: Optional[BlockType]
        Transformations to apply on the inputs after the module is
        called (so **after** `forward`).
        Defaults to None
    aggregation: Optional[TabularAggregationType]
        Aggregation to apply after processing the  `forward`-method to output a single Tensor.
        Defaults to None
    seq: bool
        Whether to process inputs for sequential model (returns 3-D tensor)
        or not (returns 2-D tensor). Use `seq=True` to treat the sparse (list) features
        as sequences (e.g. for sequential recommendation) and `seq=False` to treat sparse
        features as multi-hot categorical representations.
        Defaults to False
    add_continuous_branch: bool
        If set, add the branch to process continuous features
        Defaults to True
    continuous_tags: Optional[Union[TagsType, Tuple[Tags]]]
        Tags to filter the continuous features
        Defaults to  (Tags.CONTINUOUS,)
    continuous_projection: Optional[Block]
        If set, concatenate all numerical features and projet using the
        specified Block.
        Defaults to None
    add_embedding_branch: bool
        If set, add the branch to process categorical features
        Defaults to True
    embedding_options : EmbeddingOptions, optional
        An EmbeddingOptions instance, which allows for a number of
        options for the embedding table, by default EmbeddingOptions()
    categorical_tags: Optional[Union[TagsType, Tuple[Tags]]]
        Tags to filter the continuous features
        Defaults to (Tags.CATEGORICAL,)
    sequential_tags: Optional[Union[TagsType, Tuple[Tags]]]
        Tags to filter the sparse features
        Defaults to (Tags.SEQUENCE,)
    split_sparse: Optional[bool]
        When True, separate the processing of context (2-D) and sparse features (3-D).
        Defaults to False
    masking: Optional[Union[str, MaskSequence]], optional
        If set, Apply masking to the input embeddings and compute masked labels.
        Defaults to None
    seq_aggregator: Block
        If non-sequential model (seq=False):
        aggregate the sparse features tensor along the sequence axis.
        Defaults to SequenceAggregator('mean')
    """
    branches = branches or {}

    if split_sparse:
        sparse_schema = schema.select_by_tag(sequential_tags)
        context_schema = schema.remove_by_tag(sequential_tags)
        if not sparse_schema:
            raise ValueError(
                "Please make sure that schema has features tagged as 'sequence' when"
                "`split_context` is set to True"
            )
        if not aggregation:
            LOG.info(
                "aggregation is not provided, "
                "default `concat` will be used to merge sequential features"
            )
            aggregation = "concat"
        agg = aggregation
        sparse_interactions = InputBlock(
            sparse_schema,
            branches,
            post,
            aggregation=agg,
            seq=True,
            max_seq_length=max_seq_length,
            add_continuous_branch=add_continuous_branch,
            continuous_tags=continuous_tags,
            continuous_projection=continuous_projection,
            add_embedding_branch=add_embedding_branch,
            embedding_options=embedding_options,
            categorical_tags=categorical_tags,
            split_sparse=False,
        )
        if masking:
            if isinstance(masking, str):
                masking = masking_registry.parse(masking)()
            sparse_interactions = sparse_interactions.connect(masking)

        if not seq:
            sparse_interactions = sparse_interactions.connect(seq_aggregator)

        if not context_schema:
            return sparse_interactions

        branches["sparse"] = sparse_interactions
        return InputBlock(
            context_schema,
            branches,
            post,
            aggregation=agg,
            seq=False,
            add_continuous_branch=add_continuous_branch,
            continuous_tags=continuous_tags,
            continuous_projection=continuous_projection,
            add_embedding_branch=add_embedding_branch,
            embedding_options=embedding_options,
            categorical_tags=categorical_tags,
            split_sparse=False,
        )

    if (
        add_continuous_branch
        and schema.select_by_tag(continuous_tags).excluding_by_tag(Tags.TARGET).column_schemas
    ):
        pre = None
        if max_seq_length and seq:
            pre = AsDenseFeatures(max_seq_length)
        branches["continuous"] = ContinuousFeatures.from_schema(  # type: ignore
            schema,
            tags=continuous_tags,
            pre=pre,
        )
    if (
        add_embedding_branch
        and schema.select_by_tag(categorical_tags).excluding_by_tag(Tags.TARGET).column_schemas
    ):
        emb_cls: Type[EmbeddingFeatures] = SequenceEmbeddingFeatures if seq else EmbeddingFeatures
        emb_kwargs = {}
        if max_seq_length and seq:
            emb_kwargs["max_seq_length"] = max_seq_length

        branches["categorical"] = emb_cls.from_schema(  # type: ignore
            schema, tags=categorical_tags, embedding_options=embedding_options, **emb_kwargs
        )

    if continuous_projection:
        return ContinuousEmbedding(
            ParallelBlock(branches),
            continuous_projection,
            aggregation=aggregation,
            post=post,
            name="continuous_projection",
        )

    return ParallelBlock(branches, aggregation=aggregation, post=post, is_input=True, **kwargs)



#
# Copyright (c) 2021, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import pytest
import tensorflow as tf

import merlin.models.tf as ml
from merlin.io import Dataset
from merlin.models.config.schema import FeatureCollection
from merlin.models.tf.blocks.core.context import FeatureContext
from merlin.models.tf.blocks.core.transformations import AsDenseFeatures
from merlin.schema import Tags


@pytest.mark.parametrize("mask_block", [ml.CausalLanguageModeling, ml.MaskedLanguageModeling])
def test_masking_block(sequence_testing_data: Dataset, mask_block):

    schema_list = sequence_testing_data.schema.select_by_tag(Tags.SEQUENCE)
    embedding_block = ml.InputBlock(schema_list, aggregation="concat", max_seq_length=4, seq=True)
    block = embedding_block.connect(mask_block(), context=ml.ModelContext())

    batch = ml.sample_batch(sequence_testing_data, batch_size=100, include_targets=False)
    feature_context = FeatureContext(FeatureCollection(schema_list, AsDenseFeatures()(batch)))
    masked_input = block(batch, feature_context=feature_context)

    assert masked_input.shape[-1] == 148
    assert masked_input.shape[1] == 4


def test_mask_error(sequence_testing_data: Dataset):
    schema_list = sequence_testing_data.schema.select_by_tag(Tags.SEQUENCE)

    features = FeatureCollection(schema_list, None)
    feature_context = FeatureContext(features)

    with pytest.raises(ValueError) as excinfo:
        _ = feature_context.mask
    assert "The mask is not stored" in str(excinfo.value)


# Test only last item is masked when eval_on_last_item_seq_only
@pytest.mark.parametrize("mask_block", [ml.CausalLanguageModeling, ml.MaskedLanguageModeling])
def test_masking_only_last_item_for_eval(sequence_testing_data, mask_block):
    schema_list = sequence_testing_data.schema.select_by_tag(Tags.SEQUENCE)
    embedding_block = ml.InputBlock(schema_list, aggregation="concat", max_seq_length=4, seq=True)
    task = ml.BinaryClassificationTask("click")

    model = ml.Model(embedding_block, mask_block(), task)

    batch = ml.sample_batch(
        sequence_testing_data, batch_size=100, include_targets=False, to_dense=True
    )

    features = FeatureCollection(schema_list, model.as_dense(batch))
    feature_context = FeatureContext(features)
    _ = model(batch, training=False, feature_context=feature_context)

    # Get last non-padded label from input
    item_ids = batch["item_id_seq"]
    non_padded_mask = item_ids != 0
    rows_ids = tf.range(item_ids.shape[0], dtype=tf.int64)
    last_item_sessions = tf.reduce_sum(tf.cast(non_padded_mask, tf.int64), axis=1) - 1
    indices = tf.concat(
        [tf.expand_dims(rows_ids, 1), tf.expand_dims(last_item_sessions, 1)], axis=1
    )
    last_labels = tf.gather_nd(item_ids, indices).numpy()
    # get the mask from the model
    trgt_mask = feature_context._mask

    # check that only one item is masked for each session
    assert tf.reduce_sum(tf.cast(trgt_mask, tf.int32)).numpy() == batch["item_id_seq"].shape[0]
    # check only the last non-paded item is masked
    out_targets = tf.boolean_mask(batch["item_id_seq"], trgt_mask).numpy()
    assert all(last_labels == out_targets)


# Test at least one item is masked when training
def test_at_least_one_masked_item_mlm(sequence_testing_data):
    schema_list = sequence_testing_data.schema.select_by_tag(Tags.SEQUENCE)
    embedding_block = ml.InputBlock(schema_list, max_seq_length=4, aggregation="concat", seq=True)
    mask_block = ml.MaskedLanguageModeling()
    task = ml.BinaryClassificationTask("click")
    model = ml.Model(embedding_block, mask_block, task)

    batch = ml.sample_batch(sequence_testing_data, batch_size=100, include_targets=False)

    features = FeatureCollection(schema_list, model.as_dense(batch))
    feature_context = FeatureContext(features)
    _ = model(batch, feature_context=feature_context, training=True)

    trgt_mask = tf.cast(feature_context.mask, tf.int32)
    assert all(tf.reduce_sum(trgt_mask, axis=1).numpy() > 0)


# Check that not all items are masked when training
def test_not_all_masked_lm(sequence_testing_data):
    schema_list = sequence_testing_data.schema.select_by_tag(Tags.SEQUENCE)
    embedding_block = ml.InputBlock(schema_list, aggregation="concat", seq=True, max_seq_length=4)
    mask_block = ml.MaskedLanguageModeling()
    task = ml.BinaryClassificationTask("click")
    model = ml.Model(embedding_block, mask_block, task)

    batch = ml.sample_batch(
        sequence_testing_data, batch_size=100, include_targets=False, to_dense=True
    )
    features = FeatureCollection(schema_list, model.as_dense(batch))
    feature_context = FeatureContext(features)
    _ = model(batch, feature_context=feature_context, training=True)

    trgt_mask = tf.cast(feature_context.mask, tf.int32)
    non_padded_mask = batch["item_id_seq"] != 0
    assert all(trgt_mask.numpy().sum(axis=1) != non_padded_mask.numpy().sum(axis=1))



#
# Copyright (c) 2021, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import tensorflow as tf

import merlin.models.tf as ml


def test_fm_pairwise_interaction():
    NUM_ROWS = 100
    NUM_FEATS = 10
    EMBED_DIM = 64

    inputs = tf.random.uniform((NUM_ROWS, NUM_FEATS, EMBED_DIM))

    pairwise_interaction = ml.FMPairwiseInteraction()
    outputs = pairwise_interaction(inputs)

    assert list(outputs.shape) == [NUM_ROWS, EMBED_DIM]



#
# Copyright (c) 2021, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from typing import Optional, Sequence

import tensorflow as tf
from tensorflow.python.layers.base import Layer

from merlin.models.tf.blocks.core.base import Block
from merlin.models.tf.blocks.core.transformations import LogitsTemperatureScaler
from merlin.models.tf.blocks.retrieval.base import ItemRetrievalScorer
from merlin.models.tf.blocks.sampling.base import ItemSampler
from merlin.models.tf.blocks.sampling.in_batch import InBatchSampler
from merlin.models.tf.metrics.ranking import ranking_metrics
from merlin.models.tf.prediction_tasks.classification import MultiClassClassificationTask
from merlin.models.utils import schema_utils
from merlin.schema import Schema, Tags


@tf.keras.utils.register_keras_serializable(package="merlin_models")
class ItemRetrievalTask(MultiClassClassificationTask):
    """Prediction-task for item-retrieval.

    Parameters
    ----------
        schema: Schema
            The schema object including features to use and their properties.
        samplers: List[ItemSampler]
            List of samplers for negative sampling, by default `[InBatchSampler()]`
        post_logits: Optional[PredictionBlock]
            Optional extra pre-call block for post-processing the logits, by default None.
            You can for example use `post_logits = mm.PopularitySamplingBlock(item_fequency)`
            for populariy sampling correction.
        target_name: Optional[str]
            If specified, name of the target tensor to retrieve from dataloader.
            Defaults to None.
        task_name: Optional[str]
            name of the task.
            Defaults to None.
        task_block: Block
            The `Block` that applies additional layers op to inputs.
            Defaults to None.
        logits_temperature: float
            Parameter used to reduce the model overconfidence, so that logits / T.
            Defaults to 1.
    Returns
    -------
        PredictionTask
            The item retrieval prediction task
    """

    DEFAULT_LOSS = "categorical_crossentropy"
    DEFAULT_METRICS = ranking_metrics(top_ks=[10])

    def __init__(
        self,
        schema: Schema,
        samplers: Sequence[ItemSampler] = (),
        target_name: Optional[str] = None,
        task_name: Optional[str] = None,
        task_block: Optional[Layer] = None,
        post_logits: Optional[Block] = None,
        logits_temperature: float = 1.0,
        cache_query: bool = False,
        store_negative_ids: bool = False,
        **kwargs,
    ):
        self.item_id_feature_name = schema.select_by_tag(Tags.ITEM_ID).column_names[0]
        self.item_domain = schema_utils.categorical_domains(schema)[self.item_id_feature_name]
        self.cache_query = cache_query
        pre = self._build_prediction_call(
            samplers,
            logits_temperature,
            post_logits,
            store_negative_ids,
        )

        super().__init__(
            target_name=target_name,
            task_name=task_name,
            task_block=task_block,
            pre=pre,
            **kwargs,
        )

    def call(self, inputs, training=False, eval_sampling=False, **kwargs):
        return inputs

    def _build_prediction_call(
        self,
        samplers: Sequence[ItemSampler],
        logits_temperature: float,
        post_logits: Optional[Block] = None,
        store_negative_ids: bool = False,
        **kwargs,
    ):
        if samplers is None or len(samplers) == 0:
            samplers = (InBatchSampler(),)

        prediction_call = ItemRetrievalScorer(
            samplers=samplers,
            item_id_feature_name=self.item_id_feature_name,
            item_domain=self.item_domain,
            cache_query=self.cache_query,
            store_negative_ids=store_negative_ids,
        )

        if post_logits is not None:
            prediction_call = prediction_call.connect(post_logits)

        if logits_temperature != 1:
            prediction_call = prediction_call.connect(LogitsTemperatureScaler(logits_temperature))

        return prediction_call

    @property
    def retrieval_scorer(self):
        def find_retrieval_scorer_block(block):
            if isinstance(block, ItemRetrievalScorer):
                return block

            if getattr(block, "layers", None):
                for subblock in block.layers:
                    result = find_retrieval_scorer_block(subblock)
                    if result:
                        return result

            return None

        result = find_retrieval_scorer_block(self.pre)

        if result is None:
            raise Exception("An ItemRetrievalScorer layer was not found in the model.")

        return result

    def set_retrieval_cache_query(self, value: bool):
        self.retrieval_scorer.cache_query = value



#
# Copyright (c) 2022, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from merlin.models.config.schema import Feature, FeatureCollection
from merlin.schema import ColumnSchema, Schema, Tags


def test_select_features_by_name():
    schema = Schema(
        [
            ColumnSchema("a", tags=["a"]),
            ColumnSchema("b", tags=["a"]),
            ColumnSchema("c", tags=["c"]),
        ]
    )
    value = {"a": [1, 2, 3], "b": [3, 4, 5], "c": [6, 7, 8]}
    fc = FeatureCollection(schema, value)
    fc_b = fc.select_by_name("b")
    fc_bc = fc.select_by_name(["b", "c"])

    assert fc_b["b"].schema.name in "b"
    assert fc_b["b"].value == value["b"]

    assert fc_bc["b"].schema.name in "b"
    assert fc_bc["b"].value == value["b"]

    assert fc_bc["c"].schema.name in "c"
    assert fc_bc["c"].value == value["c"]


def test_select_features_by_tag():
    schema = Schema(
        [
            ColumnSchema("a", tags=[Tags.CATEGORICAL]),
            ColumnSchema("b", tags=[Tags.CATEGORICAL]),
            ColumnSchema("c", tags=[Tags.CONTINUOUS]),
        ]
    )

    value = {"a": [1, 2, 3], "b": [3, 4, 5], "c": [6, 7, 8]}

    features = FeatureCollection(schema, value)

    categorical = features.select_by_tag(Tags.CATEGORICAL)
    continuous = features.select_by_tag(Tags.CONTINUOUS)

    for feature_name in ["a", "b"]:
        assert categorical[feature_name].schema.name == feature_name
        assert categorical[feature_name].value == value[feature_name]

    assert continuous["c"].schema.name == "c"
    assert continuous["c"].value == value["c"]


def test_update_feature_schemas():
    schema = Schema(["a"])
    value = {"a": [1.0, 2.0]}
    features = FeatureCollection(schema, value)

    new_schema = Schema([schema.column_schemas["a"].with_tags("updated")])
    updated_features = features.with_schema(new_schema)

    assert updated_features.schema == new_schema


def test_get_feature():
    schema = Schema(["a"])
    value = {"a": [1.0, 2.0]}
    features = FeatureCollection(schema, value)

    feature = features["a"]

    assert isinstance(feature, Feature)
    assert feature.schema.name == "a"
    assert feature.value == value["a"]



#
# Copyright (c) 2021, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from typing import Optional

from merlin.models.tf.blocks.core.base import Block, Debug
from merlin.models.tf.blocks.core.combinators import Filter, ParallelBlock, SequentialBlock
from merlin.models.tf.blocks.interaction import DotProductInteraction
from merlin.models.tf.inputs.continuous import ContinuousFeatures
from merlin.models.tf.inputs.embedding import EmbeddingFeatures, EmbeddingOptions
from merlin.schema import Schema, Tags


def DLRMBlock(
    schema: Schema,
    embedding_dim: int,
    bottom_block: Optional[Block] = None,
    top_block: Optional[Block] = None,
) -> SequentialBlock:
    """Builds the DLRM architecture, as proposed in the following
    `paper https://arxiv.org/pdf/1906.00091.pdf`_ [1]_.

    References
    ----------
    .. [1] Naumov, Maxim, et al. "Deep learning recommendation model for
       personalization and recommendation systems." arXiv preprint arXiv:1906.00091 (2019).

    Parameters
    ----------
    schema : Schema
        The `Schema` with the input features
    bottom_block : Block
        The `Block` that combines the continuous features (typically a `MLPBlock`)
    top_block : Optional[Block], optional
        The optional `Block` that combines the outputs of bottom layer and of
        the factorization machine layer, by default None
    embedding_dim : Optional[int], optional
        Dimension of the embeddings, by default None

    Returns
    -------
    SequentialBlock
        The DLRM block

    Raises
    ------
    ValueError
        The schema is required by DLRM
    ValueError
        The bottom_block is required by DLRM
    ValueError
        The embedding_dim (X) needs to match the last layer of bottom MLP (Y).
    """
    if schema is None:
        raise ValueError("The schema is required by DLRM")

    if embedding_dim is None:
        raise ValueError("The embedding_dim is required")

    con_schema = schema.select_by_tag(Tags.CONTINUOUS).excluding_by_tag(Tags.TARGET)
    cat_schema = schema.select_by_tag(Tags.CATEGORICAL).excluding_by_tag(Tags.TARGET)

    if not len(cat_schema) > 0:
        raise ValueError("DLRM requires categorical features")

    if (
        embedding_dim is not None
        and bottom_block is not None
        and embedding_dim != bottom_block.layers[-1].units
    ):
        raise ValueError(
            f"The embedding_dim ({embedding_dim}) needs to match the "
            "last layer of bottom MLP ({bottom_block.layers[-1].units}) "
        )

    embeddings = EmbeddingFeatures.from_schema(
        cat_schema, embedding_options=EmbeddingOptions(embedding_dim_default=embedding_dim)
    )

    if len(con_schema) > 0:
        if bottom_block is None:
            raise ValueError(
                "The bottom_block is required by DLRM when "
                "continuous features are available in the schema"
            )
        con = ContinuousFeatures.from_schema(con_schema)
        bottom_block = con.connect(bottom_block)  # type: ignore
        interaction_inputs = ParallelBlock({"embeddings": embeddings, "bottom_block": bottom_block})
    else:
        interaction_inputs = embeddings  # type: ignore
        bottom_block = None

    interaction_inputs = interaction_inputs.connect(Debug())

    if not top_block:
        return interaction_inputs.connect(DotProductInteractionBlock())

    if not bottom_block:
        return interaction_inputs.connect(DotProductInteractionBlock(), top_block)

    top_block_inputs = interaction_inputs.connect_with_shortcut(
        DotProductInteractionBlock(),
        shortcut_filter=Filter("bottom_block"),
        aggregation="concat",
    )
    top_block_outputs = top_block_inputs.connect(top_block)

    return top_block_outputs


def DotProductInteractionBlock():
    return SequentialBlock(DotProductInteraction(), pre_aggregation="stack")



#
# Copyright (c) 2021, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import pytest

import merlin.models.torch as ml
from merlin.schema import Tags

torch_utils = pytest.importorskip("merlin.models.torch.utils.torch_utils")


def test_tabular_features(tabular_schema, torch_tabular_data):
    schema = tabular_schema
    tab_module = ml.TabularFeatures.from_schema(schema)

    outputs = tab_module(torch_tabular_data)

    con = schema.select_by_tag(Tags.CONTINUOUS).column_names
    cat = schema.select_by_tag(Tags.CATEGORICAL).column_names

    assert set(outputs.keys()) == set(con + cat)


def test_tabular_features_embeddings_options(tabular_schema, torch_tabular_data):
    schema = tabular_schema

    EMB_DIM = 100
    tab_module = ml.TabularFeatures.from_schema(schema, embedding_dim_default=EMB_DIM)

    outputs = tab_module(torch_tabular_data)

    categ_features = schema.select_by_tag(Tags.CATEGORICAL).column_names
    assert all(v.shape[-1] == EMB_DIM for k, v in outputs.items() if k in categ_features)


def test_tabular_features_with_projection(tabular_schema, torch_tabular_data):
    schema = tabular_schema
    tab_module = ml.TabularFeatures.from_schema(schema, continuous_projection=64)

    outputs = tab_module(torch_tabular_data)

    continuous_feature_names = schema.select_by_tag(Tags.CONTINUOUS).column_names

    assert len(set(continuous_feature_names).intersection(set(outputs.keys()))) == 0
    assert "continuous_projection" in outputs
    assert list(outputs["continuous_projection"].shape)[1] == 64


def test_tabular_features_soft_encoding(tabular_schema, torch_tabular_data):
    schema = tabular_schema

    emb_cardinality = 10
    emb_dim = 8
    tab_module = ml.TabularFeatures.from_schema(
        schema,
        continuous_soft_embeddings=True,
        soft_embedding_cardinality_default=emb_cardinality,
        soft_embedding_dim_default=emb_dim,
    )

    outputs = tab_module(torch_tabular_data)

    con = schema.select_by_tag(Tags.CONTINUOUS).column_names
    cat = schema.select_by_tag(Tags.CATEGORICAL).column_names

    assert set(outputs.keys()) == set(cat + con)

    assert all(
        list(outputs[col_name].shape) == list(torch_tabular_data[col_name].shape) + [emb_dim]
        for col_name in schema.select_by_tag(Tags.CONTINUOUS).column_names
    )



#
# Copyright (c) 2021, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import abc
import re

# Camel case to snake case utils
from typing import Generic, TypeVar

_first_cap_re = re.compile("(.)([A-Z][a-z0-9]+)")
_all_cap_re = re.compile("([a-z0-9])([A-Z])")


def camelcase_to_snakecase(name):
    s1 = _first_cap_re.sub(r"\1_\2", name)
    return _all_cap_re.sub(r"\1_\2", s1).lower()


def snakecase_to_camelcase(name):
    return "".join([w[0].upper() + w[1:] for w in name.split("_")])


def default_name(class_or_fn):
    """Default name for a class or function.

    This is the naming function by default for registries expecting classes or
    functions.

    Parameters
    ----------
    class_or_fn:
        class or function to be named.

    Returns
    -------
    Default name for registration.
    """
    return camelcase_to_snakecase(class_or_fn.__name__)


def default_object_name(obj):
    return default_name(type(obj))


class Registry:
    """
    Dict-like class for managing function registrations.

    Example usage::

        my_registry = Registry("custom_name")
        @my_registry.register
        def my_func():
          pass
        @my_registry.register()
        def another_func():
          pass
        @my_registry.register("non_default_name")
        def third_func(x, y, z):
          pass
        def foo():
          pass
        my_registry.register()(foo)
        my_registry.register("baz")(lambda (x, y): x + y)
        my_register.register("bar")
        print(list(my_registry))
        # ["my_func", "another_func", "non_default_name", "foo", "baz"]
        # (order may vary)
        print(my_registry["non_default_name"] is third_func)  # True
        print("third_func" in my_registry)                    # False
        print("bar" in my_registry)                           # False
        my_registry["non-existent_key"]                       # raises KeyError

    Optional validation, on_set callback and value transform also supported.

    Parameters
    ----------
    registry_name: str
        identifier for the given registry. Used in error msgs.
    default_key_fn: callable, optional
        function mapping value -> key for registration when a key is not provided
    validator: callable, optional
        if given, this is run before setting a given (key, value) pair. Accepts (key, value)
        and should raise if there is a problem. Overwriting existing keys is not allowed and is
        checked separately. Values are also checked to be callable separately.
    on_set: callable, optional
        callback function accepting (key, value) pair which is run after an item is successfully
        set.
    value_transformer: callable, optional
        if run, `__getitem__` will return value_transformer(key, registered_value).
    """

    def __init__(
        self,
        registry_name,
        default_key_fn=default_name,
        validator=None,
        on_set=None,
        value_transformer=(lambda k, v: v),
    ):
        self._registry = {}
        self._name = registry_name
        self._default_key_fn = default_key_fn
        self._validator = validator
        self._on_set = on_set
        self._value_transformer = value_transformer

    @classmethod
    def class_registry(
        cls, registry_name, default_key_fn=default_name, validator=None, on_set=None
    ):
        return cls(
            registry_name=registry_name,
            default_key_fn=default_key_fn,
            validator=validator,
            on_set=on_set,
            value_transformer=(lambda k, v: v()),
        )

    def default_key(self, value):
        """Default key used when key not provided. Uses function from __init__."""
        return self._default_key_fn(value)

    @property
    def name(self):
        return self._name

    def validate(self, key, value):
        """Validation function run before setting. Uses function from __init__."""
        if self._validator is not None:
            self._validator(key, value)

    def on_set(self, key, value):
        """Callback called on successful set. Uses function from __init__."""
        if self._on_set is not None:
            self._on_set(key, value)

    def __setitem__(self, key, value):
        """Validate, set, and (if successful) call `on_set` for the given item.

        Parameters
        ----------
        key:
            key to store value under. If `None`, `self.default_key(value)` is used.
        value:
            callable stored under the given key.
        Raises
        ------
            KeyError: if key is already in registry.
        """
        if key is None:
            key = self.default_key(value)
        if not isinstance(key, tuple):
            key = (key,)

        for k in key:
            if k in self:
                raise KeyError("key %s already registered in registry %s" % (k, self._name))
            if not callable(value):
                raise ValueError("value must be callable")
            self.validate(k, value)
            self._registry[k] = value
            self.on_set(k, value)

    def register(self, key_or_value=None):
        """Decorator to register a function, or registration itself.
        This is primarily intended for use as a decorator, either with or without
        a key/parentheses.

        Example Usage::

            @my_registry.register('key1')
            def value_fn(x, y, z):
              pass
            @my_registry.register()
            def another_fn(x, y):
              pass
            @my_registry.register
            def third_func():
              pass

        Note if key_or_value is provided as a non-callable, registration only
        occurs once the returned callback is called with a callable as its only
        argument::

            callback = my_registry.register('different_key')
            'different_key' in my_registry  # False
            callback(lambda (x, y): x + y)
            'different_key' in my_registry  # True

        Parameters
        ----------
        key_or_value (optional):
            key to access the registered value with, or the unction itself. If `None` (default),
            `self.default_key` will be called on `value` once the returned callback is called with
            `value` as the only arg. If `key_or_value` is itself callable, it is assumed to be the
            value and the key is given by `self.default_key(key)`.

        Returns
        -------
        decorated callback, or callback generated a decorated function.
        """

        def decorator(value, key):
            self[key] = value
            return value

        # Handle if decorator was used without parens
        if callable(key_or_value):
            return decorator(value=key_or_value, key=None)
        else:
            return lambda value: decorator(value, key=key_or_value)

    def register_with_multiple_names(self, *names):
        return self.register(names)

    def __getitem__(self, key):
        if key not in self:
            raise KeyError(
                "%s never registered with registry %s. Available:\n %s"
                % (key, self.name, display_list_by_prefix(sorted(self), 4))
            )
        value = self._registry[key]
        return self._value_transformer(key, value)

    def __contains__(self, key):
        return key in self._registry

    def keys(self):
        return self._registry.keys()

    def values(self):
        return (self[k] for k in self)  # complicated because of transformer

    def items(self):
        return ((k, self[k]) for k in self)  # complicated because of transformer

    def __iter__(self):
        return iter(self._registry)

    def __len__(self):
        return len(self._registry)

    def _clear(self):
        self._registry.clear()

    def get(self, key, default=None):
        return self[key] if key in self else default

    def parse(self, class_or_str):
        if isinstance(class_or_str, str):
            return self[class_or_str]

        return class_or_str


RegistryClassT = TypeVar("RegistryClassT")


class RegistryMixin(Generic[RegistryClassT], abc.ABC):
    registry: Registry

    @classmethod
    def parse(cls, class_or_str) -> RegistryClassT:
        output: RegistryClassT = cls.registry.parse(class_or_str)  # type: ignore

        return output


def display_list_by_prefix(names_list, starting_spaces=0):
    """Creates a help string for names_list grouped by prefix."""
    cur_prefix, result_lines = None, []
    space = " " * starting_spaces
    for name in sorted(names_list):
        split = name.split("_", 1)
        prefix = split[0]
        if cur_prefix != prefix:
            result_lines.append(space + prefix + ":")
            cur_prefix = prefix
        result_lines.append(space + "  * " + name)
    return "\n".join(result_lines)



#
# Copyright (c) 2021, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import abc
from typing import List, Optional

import tensorflow as tf
from tensorflow.keras.layers import Layer

from merlin.models.tf.blocks.core.base import EmbeddingWithMetadata
from merlin.models.tf.typing import TabularData


class ItemSampler(abc.ABC, Layer):
    def __init__(
        self,
        max_num_samples: Optional[int] = None,
        **kwargs,
    ):
        super(ItemSampler, self).__init__(**kwargs)
        self.set_max_num_samples(max_num_samples)

    @abc.abstractmethod
    def add(self, embeddings: tf.Tensor, items_metadata: TabularData, training=True):
        raise NotImplementedError()

    @abc.abstractmethod
    def sample(self) -> EmbeddingWithMetadata:
        raise NotImplementedError()

    def _check_inputs_batch_sizes(self, inputs: TabularData):
        embeddings_batch_size = tf.shape(inputs["embeddings"])[0]
        for feat_name in inputs["metadata"]:
            metadata_feat_batch_size = tf.shape(inputs["metadata"][feat_name])[0]

            tf.assert_equal(
                embeddings_batch_size,
                metadata_feat_batch_size,
                "The batch size (first dim) of embeddings "
                f"({int(embeddings_batch_size)}) and metadata "
                f"features ({int(metadata_feat_batch_size)}) must match.",
            )

    @property
    def required_features(self) -> List[str]:
        return []

    @property
    def max_num_samples(self) -> int:
        return self._max_num_samples

    def set_max_num_samples(self, value) -> None:
        self._max_num_samples = value



#
# Copyright (c) 2022, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from typing import Union

import numpy as np
import pandas as pd
from scipy.sparse import coo_matrix

from merlin.io import Dataset
from merlin.schema import Tags


def dataset_to_coo(dataset: Dataset):
    """Converts a merlin.io.Dataset object to a scipy coo matrix"""
    user_id_column = dataset.schema.select_by_tag(Tags.USER_ID).first.name
    item_id_column = dataset.schema.select_by_tag(Tags.ITEM_ID).first.name

    columns = [user_id_column, item_id_column]
    target_column = None
    target = dataset.schema.select_by_tag(Tags.TARGET)

    if len(target) > 1:
        raise ValueError(
            "Found more than one column tagged Tags.TARGET in the dataset schema."
            f" Expected a single target column but found  {target.column_names}"
        )

    elif len(target) == 1:
        target_column = target.first.name
        columns.append(target_column)

    df = dataset.to_ddf()[columns].compute(scheduler="synchronous")

    userids = _to_numpy(df[user_id_column])
    itemids = _to_numpy(df[item_id_column])
    targets = _to_numpy(df[target_column]) if target_column else np.ones(len(userids))
    return coo_matrix((targets.astype("float32"), (userids, itemids)))


def unique_rows_by_features(
    dataset: Dataset, features_tag: Union[str, Tags], grouping_tag: Union[str, Tags]
):
    # Check if merlin-dataset is passed
    ddf = dataset.to_ddf() if hasattr(dataset, "to_ddf") else dataset

    columns = dataset.schema.select_by_tag(features_tag).column_names
    if columns:
        id_col = dataset.schema.select_by_tag(grouping_tag).first.name
        ddf = ddf[columns].drop_duplicates(id_col, keep="first")

    return Dataset(ddf)


def _to_numpy(series):
    """converts a pandas or cudf series to a numpy array"""
    if isinstance(series, pd.Series):
        return series.values
    else:
        return series.values_host



# #
# # Copyright (c) 2021, NVIDIA CORPORATION.
# #
# # Licensed under the Apache License, Version 2.0 (the "License");
# # you may not use this file except in compliance with the License.
# # You may obtain a copy of the License at
# #
# #     http://www.apache.org/licenses/LICENSE-2.0
# #
# # Unless required by applicable law or agreed to in writing, software
# # distributed under the License is distributed on an "AS IS" BASIS,
# # WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# # See the License for the specific language governing permissions and
# # limitations under the License.
# #
#
import pytest

import merlin.models.tf as ml
from merlin.models.tf.utils import testing_utils


@pytest.mark.parametrize("run_eagerly", [True, False])
def test_ncf_model(ecommerce_data, run_eagerly):
    model = ml.benchmark.NCFModel(
        ecommerce_data.schema,
        embedding_dim=64,
        mlp_block=ml.MLPBlock([64]),
        prediction_tasks=ml.BinaryClassificationTask("click"),
    )

    testing_utils.model_test(model, ecommerce_data, run_eagerly=run_eagerly)



# Copyright (c) 2021, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


def docstring_parameter(*args, extra_padding=None, **kwargs):
    def dec(obj):
        if extra_padding:

            def pad(value):
                return ("\n" + " " * extra_padding).join(value.split("\n"))

            nonlocal args, kwargs
            kwargs = {key: pad(value) for key, value in kwargs.items()}
            args = [pad(value) for value in args]
        obj.__doc__ = obj.__doc__.format(*args, **kwargs)
        return obj

    return dec



#
# Copyright (c) 2021, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import difflib
import logging
import os
import pathlib
from pathlib import Path
from random import randint
from typing import Dict, Optional, Sequence, Tuple, Union

import numpy as np

import merlin.io
from merlin.models.utils import schema_utils
from merlin.schema import Schema, Tags
from merlin.schema.io.tensorflow_metadata import TensorflowMetadata

LOG = logging.getLogger("merlin-models")
HERE = pathlib.Path(__file__).parent
KNOWN_DATASETS: Dict[str, Path] = {
    "e-commerce": HERE / "ecommerce/small",
    "e-commerce-large": HERE / "ecommerce/large",
    "music-streaming": HERE / "entertainment" / "music_streaming",
    "social": HERE / "social",
    "testing": HERE / "testing",
    "sequence-testing": HERE / "testing" / "sequence_testing",
    "movielens-25m": HERE / "entertainment/movielens/25m",
    "movielens-1m": HERE / "entertainment/movielens/1m",
    "movielens-1m-raw-ratings": HERE / "entertainment/movielens/1m-raw/ratings/",
    "movielens-100k": HERE / "entertainment/movielens/100k",
    "criteo": HERE / "advertising/criteo/transformed",
    "aliccp": HERE / "ecommerce/aliccp/transformed",
    "aliccp-raw": HERE / "ecommerce/aliccp/raw",
}


def generate_data(
    input: Union[Schema, Path, str],
    num_rows: int,
    set_sizes: Sequence[float] = (1.0,),
    min_session_length=5,
    max_session_length=None,
    device="cpu",
) -> Union[merlin.io.Dataset, Tuple[merlin.io.Dataset, ...]]:
    """
    Generate synthetic data from a schema or one of the known datasets.

    Known fully synthetic datasets:
    - e-commerce
    - e-commerce-large
    - music-streaming
    - social
    - testing
    - sequence-testing

    Based on real datasets:
    - criteo
    - aliccp
    - aliccp-raw


    Parameters
    ----------
    input: Union[Schema, Path, str]
        The schema, path to a dataset or name of a known dataset.
    num_rows: int
        The number of rows to generate.
    set_sizes: Sequence[float], default=(1.0,)
        This parameter allows outputting multiple datasets, where each
        dataset is a subset of the original dataset.

        Example::
            train, valid = generate_data(input, 10000, (0.8, 0.2))
    min_session_length: int
        The minimum number of events in a session.
    max_session_length: int
        The maximum number of events in a session.
    device: str
        The device to use for the data generation.
        Supported values: {'cpu', 'gpu'}

    Returns
    -------
    merlin.io.Dataset
    """

    schema: Schema
    if isinstance(input, str):
        if input in KNOWN_DATASETS:
            input = KNOWN_DATASETS[input]
        elif not os.path.exists(input):
            closest_match = difflib.get_close_matches(input, KNOWN_DATASETS.keys(), n=1)
            raise ValueError(f"Unknown dataset {input}, did you mean: {closest_match[0]}?")

        schema = _get_schema(input)
    elif isinstance(input, Schema):
        schema = input
    else:
        raise ValueError(f"Unknown input type: {type(input)}")

    df = generate_user_item_interactions(
        schema, num_rows, min_session_length, max_session_length, device=device
    )

    if list(set_sizes) != [1.0]:
        num_rows = df.shape[0]
        output_datasets = []
        start_i = 0
        for set_size in set_sizes:
            num_rows_set = int(num_rows * set_size)
            end_i = start_i + num_rows_set
            set_df = df.iloc[start_i:end_i]
            start_i = end_i
            output_datasets.append(set_df)

        return tuple([merlin.io.Dataset(d, schema=schema) for d in output_datasets])

    return merlin.io.Dataset(df, schema=schema)


def generate_user_item_interactions(
    schema: Schema,
    num_interactions: int,
    min_session_length: int = 5,
    max_session_length: Optional[int] = None,
    device: str = "cpu",
):
    """
    Util function to generate synthetic data of user-item interactions from a schema object,
    it supports the generation of conditional user, item and session features.

    The schema should include a few tags:
    - `Tags.SESSION_ID` to tag the session-id feature.
    - `Tags.USER_ID` for user-id feature.
    - `Tags.ITEM_ID` for item-id feature.

    It supports both, GPU-based and CPU-based, generation.

    Parameters:
    ----------
    schema: Schema
        schema object describing the columns to generate.
    num_interactions: int
        number of interaction rows to generate.
    max_session_length: Optional[int]
        The maximum length of the multi-hot/sequence features
    min_session_length: int
        The minimum length of the multi-hot/sequence features
    device: str
        device to use for generating data.

    Returns
    -------
    data: DataFrame
       DataFrame with synthetic generated rows,
       If `cpu`, the function returns a Pandas dataframe,
       otherwise, it returns a cudf dataframe.
    """
    if device == "cpu":
        import numpy as _array
        import pandas as _frame
    else:
        import cudf as _frame
        import cupy as _array
    data = _frame.DataFrame()
    processed_cols = []
    # get session cols
    session_id_cols = list(schema.select_by_tag(Tags.SESSION_ID))
    if session_id_cols:
        session_id_col = session_id_cols[0]
        data[session_id_col.name] = _array.clip(
            _array.random.lognormal(3.0, 1.0, num_interactions).astype(_array.int32),
            1,
            session_id_col.int_domain.max,
        ).astype(str(session_id_col.dtype))

        features = list(schema.select_by_tag(Tags.SESSION).remove_by_tag(Tags.SESSION_ID))
        data = generate_conditional_features(
            data,
            features,
            session_id_col,
            min_session_length=min_session_length,
            max_session_length=max_session_length,
            device=device,
        )
        processed_cols += [f.name for f in features] + [session_id_col.name]

    # get USER cols
    user_id_cols = list(schema.select_by_tag(Tags.USER_ID))
    if user_id_cols:
        user_id_col = user_id_cols[0]
        data[user_id_col.name] = _array.clip(
            _array.random.lognormal(3.0, 1.0, num_interactions).astype(_array.int32),
            1,
            user_id_col.int_domain.max,
        ).astype(str(user_id_col.dtype))
        features = list(schema.select_by_tag(Tags.USER).remove_by_tag(Tags.USER_ID))
        data = generate_conditional_features(
            data,
            features,
            user_id_col,
            min_session_length=min_session_length,
            max_session_length=max_session_length,
            device=device,
        )
        processed_cols += [f.name for f in features] + [user_id_col.name]

    # get ITEM cols
    item_schema = schema.select_by_tag(Tags.ITEM_ID)
    if not len(item_schema) > 0:
        raise ValueError("Item ID column is required")
    item_id_col = item_schema.first

    is_list_feature = item_id_col.is_list
    if not is_list_feature:
        shape = num_interactions
    else:
        shape = (num_interactions, item_id_col.value_count.max)  # type: ignore
    tmp = _array.clip(
        _array.random.lognormal(3.0, 1.0, shape).astype(_array.int32),
        1,
        item_id_col.int_domain.max,
    ).astype(str(item_id_col.dtype))
    if isinstance(shape, int):
        data[item_id_col.name] = tmp
    else:
        data[item_id_col.name] = list(tmp)
    features = list(schema.select_by_tag(Tags.ITEM).remove_by_tag(Tags.ITEM_ID))
    data = generate_conditional_features(
        data,
        features,
        item_id_col,
        min_session_length=min_session_length,
        max_session_length=max_session_length,
        device=device,
    )
    processed_cols += [f.name for f in features] + [item_id_col.name]

    # Get remaining features
    remaining = schema.without(processed_cols)

    for feature in remaining.select_by_tag(Tags.BINARY_CLASSIFICATION):
        data[feature.name] = _array.random.randint(0, 2, num_interactions).astype(
            str(feature.dtype)
        )

    for feature in remaining.remove_by_tag(Tags.BINARY_CLASSIFICATION):
        is_int_feature = np.issubdtype(feature.dtype, np.integer)
        is_list_feature = feature.is_list
        if is_list_feature:
            data[feature.name] = generate_random_list_feature(
                feature, num_interactions, min_session_length, max_session_length, device
            )

        elif is_int_feature:
            domain = feature.int_domain
            min_value, max_value = (domain.min, domain.max) if domain else (0, 1)

            data[feature.name] = _array.random.randint(
                min_value, max_value, num_interactions, dtype=str(feature.dtype)
            )

        else:
            domain = feature.float_domain
            min_value, max_value = (domain.min, domain.max) if domain else (0.0, 1.0)

            data[feature.name] = _array.random.uniform(min_value, max_value, num_interactions)

    return data


def generate_conditional_features(
    data,
    features,
    parent_feature,
    min_session_length: int = 5,
    max_session_length: Optional[int] = None,
    device="cpu",
):
    """
    Generate features conditioned by the value of `parent_feature` feature
    """
    if device == "cpu":
        import numpy as _array
        import pandas as _frame
    else:
        import cudf as _frame
        import cupy as _array

    num_interactions = data.shape[0]
    for feature in features:
        is_int_feature = np.issubdtype(feature.dtype, np.integer)
        is_list_feature = feature.is_list

        if is_list_feature:
            data[feature.name] = generate_random_list_feature(
                feature, num_interactions, min_session_length, max_session_length, device
            )

        elif is_int_feature:
            if not feature.int_domain:
                raise ValueError(
                    "Int domain is required for conditional features, got {}".format(feature)
                )

            data[feature.name] = _frame.cut(
                data[parent_feature.name],
                feature.int_domain.max - 1,
                labels=list(range(1, feature.int_domain.max)),
            ).astype(str(feature.dtype))

        else:
            if feature.float_domain:
                _min, _max = feature.float_domain.min, feature.float_domain.max
            else:
                logging.warning(
                    "Couldn't find the float-domain for feature {}, assuming [0, 1]".format(feature)
                )
                _min, _max = 0.0, 1.0

            data[feature.name] = _array.random.uniform(_min, _max, num_interactions)

    return data


def generate_random_list_feature(
    feature,
    num_interactions,
    min_session_length: int = 5,
    max_session_length: Optional[int] = None,
    device="cpu",
):
    if device == "cpu":
        import numpy as _array
    else:
        import cupy as _array

    is_int_feature = np.issubdtype(feature.dtype, np.integer)
    if is_int_feature:
        if max_session_length:
            padded_array = []
            for _ in range(num_interactions):
                list_length = randint(min_session_length, max_session_length)
                actual_values = _array.random.randint(
                    1, feature.int_domain.max, (list_length,)
                ).astype(str(feature.dtype))

                padded_array.append(
                    _array.pad(
                        actual_values,
                        [0, max_session_length - list_length],
                        constant_values=0,
                    )
                )
            return list(_array.stack(padded_array, axis=0))
        else:
            list_length = feature.value_count.max
            return list(
                _array.random.randint(
                    1, feature.int_domain.max, (num_interactions, list_length)
                ).astype(str(feature.dtype))
            )
    else:
        if max_session_length:
            padded_array = []
            for _ in range(num_interactions):
                list_length = randint(min_session_length, max_session_length)
                actual_values = _array.random.uniform(
                    feature.float_domain.min, feature.float_domain.max, (list_length,)
                )

                padded_array.append(
                    _array.pad(
                        actual_values,
                        [0, max_session_length - list_length],
                        constant_values=0,
                    )
                )
            return list(_array.stack(padded_array, axis=0))
        else:
            list_length = feature.value_count.max
            return list(
                _array.random.uniform(
                    feature.float_domain.min,
                    feature.float_domain.max,
                    (num_interactions, list_length),
                )
            )


def _get_schema(path: Union[str, Path]) -> Schema:
    path = str(path)
    if os.path.isdir(path):
        if os.path.exists(os.path.join(path, "schema.json")):
            path = os.path.join(path, "schema.json")
        else:
            path = os.path.join(path, "schema.pbtxt")

    if path.endswith(".pb") or path.endswith(".pbtxt"):
        return TensorflowMetadata.from_proto_text_file(
            os.path.dirname(path), os.path.basename(path)
        ).to_merlin_schema()

    return schema_utils.tensorflow_metadata_json_to_schema(path)
